import dataclasses
import logging
from typing import Iterator, List, Tuple
from unittest import mock

import colorlog
import pytest

from pysen import logging_utils

SetupLoggerArgsType = Tuple[logging.Logger, int, logging.Handler]


@pytest.fixture
def mock_setup_logger() -> Iterator[List[SetupLoggerArgsType]]:
    args: List[SetupLoggerArgsType] = []

    def record(logger: logging.Logger, level: int, handler: logging.Handler) -> None:
        args.append((logger, level, handler))

    with mock.patch("pysen.logging_utils._setup_logger", side_effect=record):
        yield args


class TargetHandler(logging.Handler):
    def __init__(self) -> None:
        self.calls: List[str] = []
        super().__init__()

    def emit(self, record: logging.LogRecord) -> None:
        self.calls.append(record.msg)


def test_grouped_memory_handler() -> None:
    t_handler = TargetHandler()
    g_handler = logging_utils._GroupedMemoryHandler(target=t_handler)

    root_logger = logging.getLogger("test_logger")
    isort_logger = logging.getLogger("test_logger.isort")
    black_logger = logging.getLogger("test_logger.black")
    root_logger.addHandler(g_handler)

    root_logger.error("one")
    isort_logger.error("isort1")
    black_logger.error("black1")
    root_logger.error("two")
    isort_logger.error("isort2")
    black_logger.error("black2")

    g_handler.flush()

    assert t_handler.calls == ["black1", "black2", "isort1", "isort2", "one", "two"]

    formatter = logging.Formatter()
    g_handler.setFormatter(formatter)
    assert t_handler.formatter == formatter


def test__get_process_output_level_name() -> None:
    assert (
        logging_utils._get_process_output_level_name(logging.INFO)
        == logging_utils._PROCESS_STDOUT
    )
    assert (
        logging_utils._get_process_output_level_name(logging.ERROR)
        == logging_utils._PROCESS_STDERR
    )


@dataclasses.dataclass(frozen=True)
class ReporterLoggingUnitCondition:
    is_grouped: bool
    pretty: bool


def get_handler(logging_unit: logging_utils._LoggingUnit) -> logging.Handler:
    logging_unit.setup()
    handler = logging_utils.pysen_root_logger.handlers[0]
    # finalize to remove handler for ensuing tests
    logging_unit.finalize()
    return handler


class TestReporterLoggingUnit:
    def test_setup(self) -> None:
        def _get_handler(
            loglevel: int, is_grouped: bool, pretty: bool
        ) -> logging.Handler:
            return get_handler(
                logging_utils._LoggingUnit(
                    loglevel, is_grouped, pretty, is_process_enabled=True
                )
            )

        level = logging.INFO
        handler1 = _get_handler(level, True, True)
        assert isinstance(handler1, logging_utils._GroupedMemoryHandler)
        assert isinstance(handler1.target, logging.StreamHandler)
        assert isinstance(handler1.target.formatter, colorlog.ColoredFormatter)
        assert (
            handler1.target.formatter.log_colors == logging_utils._logging_output_colors
        )

        handler2 = _get_handler(level, True, False)
        assert isinstance(handler2, logging_utils._GroupedMemoryHandler)
        assert isinstance(handler2.target, logging.StreamHandler)
        assert handler2.target.formatter is not None
        assert handler2.target.formatter._fmt == "%(message)s"

        handler3 = _get_handler(level, False, True)
        assert isinstance(handler3, logging.StreamHandler)
        assert isinstance(handler3.formatter, colorlog.ColoredFormatter)
        assert handler3.formatter.log_colors == logging_utils._logging_output_colors

        handler4 = _get_handler(level, False, False)
        assert isinstance(handler4, logging.StreamHandler)
        assert handler4.formatter is not None
        assert handler4.formatter._fmt == "%(message)s"

    def test_setup_finalize(self) -> None:
        level = logging.DEBUG
        is_grouped = True
        pretty = True
        is_enabled = True

        root_logger = logging_utils.pysen_root_logger
        unit = logging_utils._LoggingUnit(level, is_grouped, pretty, is_enabled)

        unit.setup()
        assert logging_utils.reporter_root_logger.level == level
        assert logging_utils.process_output_root_logger.level == logging.INFO

        assert len(root_logger.handlers) == 1
        handler = root_logger.handlers[0]
        assert isinstance(handler, logging_utils._GroupedMemoryHandler)
        assert isinstance(handler.target, logging.StreamHandler)
        assert handler.level == level

        reporter_logger = logging_utils.get_reporter_logger("hoge")
        process_logger = logging_utils.get_process_output_logger("hoge")
        process2_logger = logging_utils.get_process_output_logger("fuga")

        reporter_logger.info("foo")
        assert len(handler._named_records) == 1
        process_logger.info("bar")
        assert len(handler._named_records) == 1
        process2_logger.info("baz")
        assert len(handler._named_records) == 2

        unit.finalize()
        # the buffer should be empty because we flushed
        assert len(handler._named_records) == 0
        assert len(root_logger.handlers) == 0



import contextlib
import pathlib
import tempfile
from typing import Iterator, Set

import git
import pytest
import tomlkit

from pysen.path import change_dir
from pysen.source import (
    PythonFileFilter,
    Source,
    SourceEntrySetting,
    _resolve,
    extension_filter,
)

BASE_DIR = pathlib.Path(__file__).resolve().parent


def test__resolve() -> None:
    current_file = pathlib.Path(__file__)
    setting = SourceEntrySetting(glob=False)
    assert _resolve(BASE_DIR, current_file.name, setting) == [current_file.absolute()]
    assert _resolve(BASE_DIR, current_file.absolute(), setting) == [
        current_file.absolute()
    ]
    assert _resolve(BASE_DIR, str(current_file.name), setting) == [
        current_file.absolute()
    ]
    assert _resolve(BASE_DIR, str(current_file.absolute()), setting) == [
        current_file.absolute()
    ]

    override_base = pathlib.Path("/opt/pysen")
    setting = SourceEntrySetting(glob=False, base_dir=override_base)
    assert _resolve(BASE_DIR, current_file.name, setting) == [
        override_base / current_file.name
    ]
    assert _resolve(BASE_DIR, current_file.absolute(), setting) == [
        current_file.absolute()
    ]
    assert _resolve(BASE_DIR, str(current_file.name), setting) == [
        override_base / current_file.name
    ]
    assert _resolve(BASE_DIR, str(current_file.absolute()), setting) == [
        current_file.absolute()
    ]

    setting = SourceEntrySetting(glob=True)

    with pytest.raises(RuntimeError):
        _resolve(BASE_DIR, pathlib.Path("**/*"), setting)

    resolved = _resolve(BASE_DIR, "**/*.py", setting)
    assert len(resolved) > 1
    assert all(str(x).startswith(str(BASE_DIR)) and x.suffix == ".py" for x in resolved)

    # NOTE(igarashi): check base_dir of glob is overridden by setting
    setting = SourceEntrySetting(glob=True, base_dir=BASE_DIR / "fakes")
    resolved_with_base = _resolve(BASE_DIR, "**/*.py", setting)
    assert len(resolved_with_base) > 1
    assert len(resolved) > len(resolved_with_base)
    assert all(
        str(x).startswith(str(BASE_DIR / "fakes")) and x.suffix == ".py"
        for x in resolved_with_base
    )

    # test if _resolve doesn't raise an error though we specify an invalid pattern
    resolved = _resolve(BASE_DIR, "hoge!%&#'/+*hoge", setting)
    assert len(resolved) == 0

    # Test if _resolve doesn't raise an error though we use a string-like object.
    # This case is intended to check that _resolve converts a string-like object
    # to the Python string object since pathlib.glob(x) now raises an error
    # if x is not a Python string object due to sys.intern(x).
    str_like = tomlkit.item("build")
    assert isinstance(str_like, str)
    assert type(str_like) is not str
    resolved = _resolve(BASE_DIR, str_like, setting)
    assert len(resolved) == 0


def test_extension_filter() -> None:
    python_file = pathlib.Path("python.py")
    cython_file = pathlib.Path("cython.pyx")
    stub_file = pathlib.Path("stub.pyi")
    markdown_file = pathlib.Path("markdown.md")
    text_file = pathlib.Path("text.txt")
    c_file = pathlib.Path("source.c")
    cpp_file = pathlib.Path("source.cpp")
    header_file = pathlib.Path("header.h")

    source = [
        python_file,
        cython_file,
        stub_file,
        markdown_file,
        text_file,
        c_file,
        cpp_file,
        header_file,
    ]

    assert [x for x in source if PythonFileFilter(x)] == [python_file, stub_file]

    predicate = extension_filter({".cpp", ".c", ".h"})
    assert [x for x in source if predicate(x)] == [c_file, cpp_file, header_file]


def test_add_remove() -> None:
    source = Source(includes=["hoge", "fuga"], excludes=["piyo"])

    with pytest.raises(KeyError):
        source.remove_include("foo")
    with pytest.raises(KeyError):
        source.remove_exclude("bar")

    source.add_include("foo")
    source.add_include("foo")
    source.remove_include("foo")
    with pytest.raises(KeyError):
        source.remove_include("foo")

    source.add_exclude("bar")
    source.add_exclude("bar")
    source.remove_exclude("bar")
    with pytest.raises(KeyError):
        source.remove_exclude("bar")

    source.remove_include("hoge")
    source.remove_include("fuga")
    source.remove_exclude("piyo")


def touch_file(path: pathlib.Path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()


@contextlib.contextmanager
def create_git_repository(
    tracked_files: Set[str], untracked_files: Set[str]
) -> Iterator[pathlib.Path]:
    files = tracked_files.union(untracked_files)
    assert len(files) == len(tracked_files) + len(
        untracked_files
    )  # assert intersection is empty

    with tempfile.TemporaryDirectory() as d:
        base_dir = pathlib.Path(d)
        with change_dir(base_dir):
            repo = git.Repo.init()
            for f in files:
                touch_file(pathlib.Path(f))

            repo.index.add(list(tracked_files))
            yield base_dir


def test_resolve_files_root_dot() -> None:
    tracked_files = {
        "A/0.md",
        "A/1.py",
        "A/stubs/0.pyi",
        "A/stubs/1.md",
        "A/stubs/2.md",
        "0.py",
        "1.md",
    }
    untracked_files = {
        "A/2.md",
        "A/3.py",
        "A/stubs/3.pyi",
        "A/stubs/4.md",
        "A/stubs/5.md",
        "2.py",
        "3.md",
    }

    source = Source()
    source.add_include(".")

    with create_git_repository(tracked_files, untracked_files) as base_dir:
        expected_tracked = {
            base_dir / "A/1.py",
            base_dir / "A/stubs/0.pyi",
            base_dir / "0.py",
        }
        expected_untracked = {
            base_dir / "A/3.py",
            base_dir / "A/stubs/3.pyi",
            base_dir / "2.py",
        }
        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)


def test_resolve_files_include_file() -> None:
    tracked_files = {
        "A/0.md",
        "A/1.py",
        "0.py",
        "1.md",
    }
    untracked_files = {
        "A/2.md",
        "A/3.py",
        "2.py",
        "3.md",
    }

    source = Source()
    source.add_include("A")

    with create_git_repository(tracked_files, untracked_files) as base_dir:
        expected_tracked = {
            base_dir / "A/1.py",
        }
        expected_untracked = {
            base_dir / "A/3.py",
        }
        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)

        source.add_include("1.md")
        source.add_include("3.md")

        expected_tracked.add(base_dir / "1.md")
        expected_untracked.add(base_dir / "3.md")

        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)

        source.add_include("**/*.md", glob=True)

        expected_tracked.add(base_dir / "A/0.md")
        expected_untracked.add(base_dir / "A/2.md")

        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)


def test_resolve_files_exclude() -> None:
    tracked_files = {
        "A/1.py",
        "A/third_party/1.py",
        "A/third_party/nested/1.py",
        "A/third_party/2.md",
        "0.py",
        "1.md",
    }
    untracked_files = {
        "A/2.py",
        "A/third_party/3.py",
        "A/third_party/nested/2.py",
        "A/third_party/4.md",
        "2.py",
        "3.md",
    }

    source = Source()
    source.add_include(".")

    with create_git_repository(tracked_files, untracked_files) as base_dir:
        expected_tracked = {
            base_dir / "A/1.py",
            base_dir / "A/third_party/1.py",
            base_dir / "A/third_party/nested/1.py",
            base_dir / "0.py",
        }
        expected_untracked = {
            base_dir / "A/2.py",
            base_dir / "A/third_party/3.py",
            base_dir / "A/third_party/nested/2.py",
            base_dir / "2.py",
        }
        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)

        source.add_exclude("0.py")
        source.add_exclude("2.py")

        expected_tracked.remove(base_dir / "0.py")
        expected_untracked.remove(base_dir / "2.py")

        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)

        source.add_exclude("A/third_party/**/*.py", glob=True)

        expected_tracked = {
            base_dir / "A/1.py",
        }
        expected_untracked = {
            base_dir / "A/2.py",
        }
        assert (
            source.resolve_files(base_dir, PythonFileFilter, use_git=True)
            == expected_tracked
        )
        assert source.resolve_files(
            base_dir, PythonFileFilter, use_git=False
        ) == expected_tracked.union(expected_untracked)


def test_copy() -> None:
    source = Source()
    source.add_include("hoge", glob=True, base_dir=pathlib.Path("/hoge"))
    source.add_include("fuga", glob=False)
    source.add_exclude("piyo", glob=True, base_dir=pathlib.Path("/fuga"))

    copied = source.copy()

    assert source.includes == copied.includes
    assert source.includes is not copied.includes
    assert source.excludes == copied.excludes
    assert source.excludes is not copied.excludes

    source.add_include("foo", glob=True)
    copied.add_exclude("bar", glob=False)

    assert source.includes.keys() == {"hoge", "fuga", "foo"}
    assert copied.includes.keys() == {"hoge", "fuga"}
    assert copied.includes == {
        "hoge": SourceEntrySetting(glob=True, base_dir=pathlib.Path("/hoge")),
        "fuga": SourceEntrySetting(glob=False),
    }

    assert source.excludes.keys() == {"piyo"}
    assert copied.excludes.keys() == {"piyo", "bar"}
    assert copied.excludes == {
        "piyo": SourceEntrySetting(glob=True, base_dir=pathlib.Path("/fuga")),
        "bar": SourceEntrySetting(glob=False),
    }



import pathlib

import pytest

from pysen.ext.mypy_wrapper import MypyPlugin, MypyTarget, run
from pysen.reporter import Reporter

BASE_DIR = pathlib.Path(__file__).resolve().parent


def test_run_zero_source() -> None:
    reporter = Reporter("mypy")
    assert run(reporter, BASE_DIR, BASE_DIR, MypyTarget([]), True) == 0


def test_mypy_plugin() -> None:
    script_plugin = MypyPlugin(script=pathlib.Path("/foo/bar/baz"))
    script_plugin2 = MypyPlugin(script=pathlib.Path("./bar/baz"))
    function_plugin = MypyPlugin(function="module_x")
    function_plugin2 = MypyPlugin(function="module_x:entry")

    with pytest.raises(ValueError):
        MypyPlugin()

    with pytest.raises(ValueError):
        MypyPlugin(script=pathlib.Path("."), function="module_y")

    assert script_plugin.as_config() == "/foo/bar/baz"
    assert script_plugin2.as_config() == "bar/baz"
    assert function_plugin.as_config() == "module_x"
    assert function_plugin2.as_config() == "module_x:entry"

    base_dir = pathlib.Path("/foo")
    assert function_plugin.as_config(base_dir) == "module_x"
    assert function_plugin2.as_config(base_dir) == "module_x:entry"
    assert script_plugin.as_config(base_dir) == "bar/baz"
    assert script_plugin2.as_config(base_dir) == "bar/baz"

    base_dir = pathlib.Path("/hoge")
    assert script_plugin.as_config(base_dir) == "../foo/bar/baz"
    assert script_plugin2.as_config(base_dir) == "bar/baz"



import pathlib

import pytest

from pysen.plugin import PluginBase
from pysen.plugin_loader import load_from_file, load_from_module, load_plugin
from pysen.pyproject_model import Config, PluginConfig

CURRENT_FILE = pathlib.Path(__file__).resolve()
BASE_DIR = CURRENT_FILE.parent


def test_load_from_file() -> None:
    root = Config()
    config = PluginConfig(
        location="tool.pysen.plugin.hoge", script=BASE_DIR / "fakes/plugin.py"
    )

    plugin = load_from_file(BASE_DIR / "fakes/plugin.py")
    assert isinstance(plugin, PluginBase)
    components = plugin.load(CURRENT_FILE, config, root)
    assert len(components) == 0

    config.config = {"enable_c1": True}
    components = plugin.load(CURRENT_FILE, config, root)
    assert len(components) == 1
    assert components[0].name == "plugin_component1"


def test_load_from_module() -> None:
    root = Config()
    config = PluginConfig(
        location="tool.pysen.plugin.hoge", function="fakes.plugin::plugin"
    )

    plugin = load_from_module("fakes.plugin::plugin")
    assert isinstance(plugin, PluginBase)
    components = plugin.load(CURRENT_FILE, config, root)
    assert len(components) == 0

    plugin = load_from_module("fakes.plugin::create")
    assert isinstance(plugin, PluginBase)
    components = plugin.load(CURRENT_FILE, config, root)
    assert len(components) == 0

    config.config = {"enable_c2": True}
    components = plugin.load(CURRENT_FILE, config, root)
    assert len(components) == 1
    assert components[0].name == "create_component2"


def test_load_plugin() -> None:
    def assert_plugin_entry_point(plugin: PluginBase, entry_point: str) -> None:
        root = Config()
        config = PluginConfig(
            location="tool.pysen.plugin.hoge", config={"enable_c1": True}
        )
        components = plugin.load(CURRENT_FILE, config, root)
        assert len(components) == 1
        assert components[0].name == f"{entry_point}_component1"

    plugin = load_plugin(function="fakes.plugin::create")
    assert isinstance(plugin, PluginBase)
    assert_plugin_entry_point(plugin, "create")

    plugin = load_plugin(script=BASE_DIR / "fakes/plugin.py")
    assert isinstance(plugin, PluginBase)
    assert_plugin_entry_point(plugin, "plugin")

    with pytest.raises(TypeError):
        load_plugin(
            function="fakes.plugin::unknown", script=BASE_DIR / "fakes/plugin.py"
        )

    with pytest.raises(TypeError):
        load_plugin()



import pathlib
import re
import tempfile
from typing import Iterator

import pytest
import tomlkit

from pysen import dumper
from pysen.setting import SettingFile

MULTIPLE_SPACES = re.compile(r"\s+")
TOOL_HEADING = "[tool]\n"


def normalize_spaces(s: str) -> str:
    return MULTIPLE_SPACES.sub(" ", s)


@pytest.fixture
def test_data() -> SettingFile:
    foo = {
        "hoge": True,
        "fuga": ["c", "A", "B", "d"],
        "piyo": "description",
    }
    bar = {
        "c": "yes",
        "a": 1.0,
        "b": 12345,
    }

    s = SettingFile()
    s.set_section(["foo"], foo)
    s.set_section(["bar"], bar)
    return s


@pytest.fixture
def test_nested_data() -> SettingFile:
    foo = {
        "piyo": "description",
        "hoge": True,
        "fuga": ["c", "A", "B", "d"],
        "hogehoge": {"z": ["1", "3", "2"], "x": "Hello", "y": 42},
    }
    bar = {
        "c": "yes",
        "a": 1.0,
        "b": 12345,
        "d": {"d-2": "hoge", "d-1": {"d-1-b": True, "d-1-a": False, "d-1-c": True}},
    }

    s = SettingFile()
    s.set_section(["tool", "foo"], foo)
    s.set_section(["tool", "bar"], bar)
    return s


@pytest.fixture
def temp_file() -> Iterator[pathlib.Path]:
    with tempfile.NamedTemporaryFile() as tmp:
        yield pathlib.Path(tmp.name)


def _assert_toml(path: pathlib.Path, expected: SettingFile) -> None:
    with path.open("r") as f:
        data = dict(tomlkit.loads(f.read()))

    assert data == expected.as_dict()


def test_dump_toml(temp_file: pathlib.Path, test_nested_data: SettingFile) -> None:
    dumper.dump_toml(temp_file, test_nested_data)
    expected = """[tool.bar] # automatically generated by pysen
# pysen ignores and overwrites any modifications
a = 1.0
b = 12345
c = "yes"
[tool.bar.d]
d-2 = "hoge"
[tool.bar.d.d-1]
d-1-a = false
d-1-b = true
d-1-c = true
[tool.foo] # automatically generated by pysen
# pysen ignores and overwrites any modifications
fuga = ["A", "B", "c", "d"]
hoge = true
piyo = "description"
[tool.foo.hogehoge]
    x = "Hello"
    y = 42
    z = ["1", "2", "3"]
"""

    with temp_file.open("r") as f:
        actual = f.read()

    # NOTE: tomlkit 0.10+ won't create a super table if the table has only one child
    if actual.startswith(TOOL_HEADING):
        actual = actual[len(TOOL_HEADING) :]
    assert normalize_spaces(actual) == normalize_spaces(expected)


def test_dump_cfg(temp_file: pathlib.Path, test_data: SettingFile) -> None:
    dumper.dump_cfg(temp_file, test_data)
    expected = """[bar]
# automatically generated by pysen
# pysen ignores and overwrites any modifications
a = 1.0
b = 12345
c = yes
[foo]
# automatically generated by pysen
# pysen ignores and overwrites any modifications
fuga = A,B,c,d
hoge = True
piyo = description
"""
    with temp_file.open("r") as f:
        actual = f.read()

    assert normalize_spaces(actual) == normalize_spaces(expected)


def test_dump(test_data: SettingFile) -> None:
    with tempfile.TemporaryDirectory() as tmp:
        d = pathlib.Path(tmp)
        dumper.dump(d, "hoge.toml", test_data)
        dumper.dump(d, "hoge.cfg", test_data)

        with pytest.raises(RuntimeError):
            dumper.dump(d, "hoge.bin", test_data)



from pathlib import Path

import pytest

from pysen.diagnostic import Diagnostic, FLCMFormatter, _format_diagnostic_position


def test_diagnostic_post_init() -> None:
    path = Path("/path/to/file")
    Diagnostic(file_path=path, message="error")
    Diagnostic(file_path=path, diff="diff")
    with pytest.raises(ValueError):
        Diagnostic(file_path=path)


def test_diagnostic_formatter() -> None:
    path = Path("/path/to/file")
    formatter = FLCMFormatter
    err = formatter.format(
        Diagnostic(
            start_line=10,
            end_line=12,
            start_column=3,
            file_path=path,
            message="line1\nline2",
        ),
        "my_command",
    )
    assert err == f"{path.resolve()}:10:3:my_command: line1\\nline2"

    err = formatter.format(
        Diagnostic(
            start_line=10,
            end_line=12,
            start_column=3,
            file_path=path,
            diff="-line1\n+line2\n",
        ),
        "my_command2",
    )
    assert err == f"{path.resolve()}:10:3:my_command2: -line1\\n+line2\\n"


def test__format_diagnostic_position() -> None:
    path = Path("/path/to/file")
    position = _format_diagnostic_position(
        Diagnostic(start_line=10, end_line=12, start_column=3, file_path=path, diff="")
    )
    assert position == f"{path.resolve()}:10:3"

    position = _format_diagnostic_position(Diagnostic(file_path=path, diff=""))
    assert position == f"{path.resolve()}:1:1"



from importlib.metadata import PackageNotFoundError
from pathlib import Path
from unittest import mock

import pytest

from pysen.exceptions import (
    DistributionNotFound,
    IncompatibleVersionError,
    UnexpectedErrorFormat,
)
from pysen.ext.isort_wrapper import (
    IsortSectionName,
    IsortSetting,
    _check_version_compatibility,
    _get_isort_version,
    _parse_file_path,
)
from pysen.py_version import VersionRepresentation


def test_export() -> None:
    setting = IsortSetting(
        known_third_party={"alpha", "beta"},
        default_section=IsortSectionName.THIRDPARTY,
        sections=[IsortSectionName.FUTURE, IsortSectionName.STDLIB],
        force_single_line=True,
        line_length=80,
    )

    name, section = setting.export()
    assert name == ["tool", "isort"]

    assert section == {
        "known_third_party": ["alpha", "beta"],
        "default_section": "THIRDPARTY",
        "sections": ["FUTURE", "STDLIB"],
        "line_length": 80,
        "force_single_line": True,
        # NOTE(igarashi): the following values are emitted by default
        "force_grid_wrap": 0,
        "include_trailing_comma": True,
        "multi_line_output": 3,
        "use_parentheses": True,
    }


def test_to_black_compatible() -> None:
    setting = IsortSetting(
        force_single_line=False,
        include_trailing_comma=False,
        multi_line_output=1,
        ensure_newline_before_comments=False,
        force_grid_wrap=1,
        use_parentheses=False,
    )
    with mock.patch(
        "pysen.ext.isort_wrapper._get_isort_version",
        return_value=VersionRepresentation(5, 0, 0),
    ):
        black_compat = setting.to_black_compatible()
    assert black_compat.multi_line_output == 3
    assert black_compat.include_trailing_comma
    assert black_compat.force_grid_wrap == 0
    assert black_compat.use_parentheses
    assert black_compat.ensure_newline_before_comments


def test__check_version_compatibility() -> None:
    with pytest.raises(IncompatibleVersionError):
        _check_version_compatibility(True, VersionRepresentation(4, 0))
    with pytest.raises(IncompatibleVersionError):
        _check_version_compatibility(False, VersionRepresentation(4, 0))
    _check_version_compatibility(None, VersionRepresentation(4, 0))
    _check_version_compatibility(True, VersionRepresentation(5, 0))
    _check_version_compatibility(False, VersionRepresentation(5, 0))
    _check_version_compatibility(None, VersionRepresentation(5, 0))


def test__parse_file_path() -> None:
    isort_format_before = (
        "/path/to/error_line_parser.py:before      2020-06-01 16:18:40.123155"
    )
    isort_format_after = (
        "/path/to/error_line_parser.py:after      2020-06-01 16:18:40.123155"
    )
    isort_format_invalid = (
        "/path/to/error_line_parser.py      2020-06-01 16:18:40.123155"
    )

    assert _parse_file_path(isort_format_before) == Path(
        "/path/to/error_line_parser.py"
    )
    assert _parse_file_path(isort_format_after) == Path("/path/to/error_line_parser.py")
    with pytest.raises(UnexpectedErrorFormat):
        _parse_file_path(isort_format_invalid)


def test__get_isort_version() -> None:
    def get_version() -> VersionRepresentation:
        _get_isort_version.cache_clear()
        return _get_isort_version()

    distro = "pysen.dist_version.distribution"
    # pass case
    with mock.patch(distro, return_value=mock.Mock(version="4.3.21")):
        assert get_version() == VersionRepresentation(4, 3, 21)
    with mock.patch(distro, return_value=mock.Mock(version="5.1.2")):
        assert get_version() == VersionRepresentation(5, 1, 2)
    # incompatible version
    with pytest.raises(IncompatibleVersionError) as e:
        with mock.patch(distro, return_value=mock.Mock(version="3.0.0")):
            get_version()
    assert "version 3.0.0 is not supported" in str(e)
    # isort cannot be imported
    with pytest.raises(DistributionNotFound) as e:
        with mock.patch(distro, side_effect=PackageNotFoundError("req", "requires")):
            get_version()
    assert "Expected isort to be installed" in str(e)



import pathlib
import subprocess

import pytest

from pysen.path import change_dir

TARGET_EXAMPLE = "simple_package"


@pytest.mark.examples
def test_cli_run(example_dir: pathlib.Path) -> None:
    target = example_dir / TARGET_EXAMPLE
    with change_dir(target):
        subprocess.run(["pysen", "run", "lint"], check=True)
        subprocess.run(["pysen", "run", "format"], check=True)
        subprocess.run(["pysen", "run", "--error-format", "gnu", "lint"], check=True)



import dataclasses
import pathlib
import tempfile
from typing import Callable, Iterator, Optional

import dacite
import pytest

from pysen import pyproject
from pysen.black import Black
from pysen.exceptions import InvalidConfigurationError
from pysen.isort import Isort, IsortSectionName
from pysen.manifest import Manifest
from pysen.mypy import Mypy, MypyFollowImports, MypyPreset
from pysen.path import change_dir
from pysen.pyproject_model import Config, LintConfig
from pysen.source import Source

FILE_PATH = pathlib.Path(__file__).resolve()
BASE_DIR = FILE_PATH.parent
ROOT_DIR = BASE_DIR.parent

PYPROJECT_PATH = ROOT_DIR / "pyproject.toml"
EXAMPLE_DIR = ROOT_DIR / "examples/simple_package"
EXAMPLE_PYPROJECT_PATH = EXAMPLE_DIR / "pyproject.toml"

assert PYPROJECT_PATH.exists()
assert EXAMPLE_DIR.exists()
assert EXAMPLE_PYPROJECT_PATH.exists()


@pytest.fixture
def temp_dir() -> Iterator[pathlib.Path]:
    with tempfile.TemporaryDirectory() as td:
        yield pathlib.Path(td)


def test_find_pyproject() -> None:
    with change_dir(BASE_DIR):
        assert pyproject.find_pyproject() == PYPROJECT_PATH

    with change_dir(BASE_DIR / "fakes/configs"):
        assert pyproject.find_pyproject() == PYPROJECT_PATH

    with change_dir(EXAMPLE_DIR):
        assert pyproject.find_pyproject() == EXAMPLE_PYPROJECT_PATH

    with pytest.raises(FileNotFoundError):
        pyproject.find_pyproject(BASE_DIR)

    with pytest.raises(FileNotFoundError):
        pyproject.find_pyproject(BASE_DIR / "no_such_file")

    assert pyproject.find_pyproject(FILE_PATH) == FILE_PATH


def test_load_manifest() -> None:
    manifest = pyproject.load_manifest(BASE_DIR / "fakes/configs/example.toml")
    assert manifest is not None
    assert isinstance(manifest, Manifest)
    components = manifest.components
    assert {x.name for x in components} == {
        "black",
        "isort",
        "mypy",
        "create_component1",
        "create_component2",
    }
    mypy = manifest.get_component("mypy")
    assert isinstance(mypy, Mypy)
    assert mypy.setting.mypy_path == [BASE_DIR / "fakes/configs/pysen-stubs"]

    assert mypy.setting.plugins is not None
    assert len(mypy.setting.plugins) == 2
    assert mypy.setting.plugins[0].script == BASE_DIR / "fakes/configs" / pathlib.Path(
        "./sugoi/plugin"
    )
    assert mypy.setting.plugins[1].function == "sugoi_plugin:entry"

    assert mypy.setting._pysen_convert_abspath

    module_settings = mypy.module_settings
    assert len(module_settings) == 4
    assert module_settings.keys() == {"pysen.stubs", "pysen.proto", "apple", "banana"}
    assert module_settings["pysen.stubs"].ignore_errors
    assert module_settings["pysen.proto"].ignore_errors
    assert not module_settings["apple"].ignore_errors
    assert module_settings["banana"].ignore_errors

    entry_setting = MypyPreset.ENTRY.get_setting(
        follow_imports=MypyFollowImports.SILENT
    )
    entry_setting._pysen_convert_abspath = True
    assert module_settings["apple"] == entry_setting

    isort = manifest.get_component("isort")
    assert isinstance(isort, Isort)
    assert isort.setting.default_section == IsortSectionName.THIRDPARTY

    manifest = pyproject.load_manifest(BASE_DIR / "fakes/configs/simple_source.toml")
    assert manifest is not None
    assert isinstance(manifest, Manifest)
    components = manifest.components
    assert len(components) == 1
    assert {x.name for x in components} == {"mypy"}

    manifest = pyproject.load_manifest(BASE_DIR / "fakes/configs/builder.toml")
    assert manifest is not None
    assert isinstance(manifest, Manifest)
    components = manifest.components
    assert len(components) == 2
    assert {x.name for x in components} == {"flake8", "isort"}


def test_resolve_lint_config_inheritance() -> None:
    source = Source(includes=[BASE_DIR])
    success = LintConfig(
        base=pathlib.Path(BASE_DIR / "fakes/configs/base.toml"),
        isort_known_first_party=["alpha"],
        source=source,
    )
    config = pyproject.resolve_lint_config_inheritance(success)
    assert config.isort_known_third_party == ["fuga", "piyo"]  # override by base.toml
    assert config.isort_known_first_party == ["alpha"]  # override by us
    assert config.line_length == 88  # override by base.toml

    failure_empty = LintConfig(
        base=pathlib.Path(BASE_DIR / "fakes/configs/empty.toml"),
        isort_known_first_party=["alpha"],
        source=source,
    )
    with pytest.raises(InvalidConfigurationError) as ex:
        pyproject.resolve_lint_config_inheritance(failure_empty)
    assert "doesn't have [tool.pysen.lint] section." in str(ex.value)


def test_load_lint_components() -> None:
    source = Source(includes=[BASE_DIR])
    success = LintConfig(
        base=pathlib.Path(BASE_DIR / "fakes/configs/base.toml"),
        isort_known_first_party=["alpha"],
        source=source,
    )
    components = pyproject.load_lint_components(success)
    assert len(components) == 2
    isort = next(x for x in components if isinstance(x, Isort))  # from base.toml
    black = next(x for x in components if isinstance(x, Black))  # from base2.toml
    assert isort is not None
    assert black is not None
    assert isort.setting.known_third_party == {"fuga", "piyo"}  # override by base.toml
    assert isort.setting.known_first_party == {"alpha"}  # override by us
    assert black.setting.line_length == 88  # override by base.toml

    failure_empty = LintConfig(
        base=pathlib.Path(BASE_DIR / "fakes/configs/empty.toml"),
        isort_known_first_party=["alpha"],
        source=source,
    )
    with pytest.raises(InvalidConfigurationError) as ex:
        pyproject.load_lint_components(failure_empty)
    assert "doesn't have [tool.pysen.lint] section." in str(ex.value)


def test_resolve_inheritance() -> None:
    @dataclasses.dataclass
    class _Model:
        message: str
        base: Optional[str] = None

    def selector(name: str) -> Callable[[pathlib.Path, Config], _Model]:
        section_path = f"tool.pysen.plugin.{name}"

        def impl(path: pathlib.Path, root: Config) -> _Model:
            config: Optional[_Model] = None
            if root.plugin is not None:
                target = next(
                    (x for x in root.plugin if x.location == section_path), None
                )
                if target is not None and target.config is not None:
                    config = dacite.from_dict(
                        _Model, target.config, dacite.Config(strict=True)
                    )
                    assert isinstance(config, _Model)

            if config is None:
                raise RuntimeError(f"doesn't have [{section_path}] section")

            return config

        return impl

    def base_selector(path: pathlib.Path, config: _Model) -> Optional[pathlib.Path]:
        if config.base is None:
            return None
        return BASE_DIR / "fakes/configs" / config.base

    def updater(lhs: _Model, rhs: _Model) -> _Model:
        lhs.message = f"{lhs.message} / {rhs.message}"
        return lhs

    config = pyproject.resolve_inheritance(
        BASE_DIR / "fakes/configs/base.toml", selector("ok"), base_selector, updater
    )
    assert config.message == "hello ok from base2 / hello ok from base"

    with pytest.raises(RuntimeError) as ex:
        pyproject.resolve_inheritance(
            BASE_DIR / "fakes/configs/base.toml",
            selector("error-nosection"),
            base_selector,
            updater,
        )

    assert "doesn't have [tool.pysen.plugin.error-nosection]" in str(ex.value)

    with pytest.raises(FileNotFoundError):
        pyproject.resolve_inheritance(
            BASE_DIR / "fakes/configs/base.toml",
            selector("error-nofile"),
            base_selector,
            updater,
        )

    base3 = BASE_DIR / "fakes/configs/base3.toml"
    with pytest.raises(InvalidConfigurationError) as e:
        pyproject.resolve_inheritance(base3, selector("ok"), base_selector, updater)
    assert f"Circular dependency detected. {base3} was visited more than once." in str(
        e
    )


def test__check_section_exists(temp_dir: pathlib.Path) -> None:
    test_file = temp_dir / "foo.toml"
    test_file.write_text("x = 0")
    assert not pyproject._check_section_exists(test_file)

    test_file.write_text(
        """[tool]
x = 0"""
    )
    assert not pyproject._check_section_exists(test_file)

    test_file.write_text(
        """[tool.pysen]
x = 0"""
    )
    assert pyproject._check_section_exists(test_file)

    test_file.write_text(
        """[tool.pysen2]
x = 0"""
    )
    assert not pyproject._check_section_exists(test_file)

    test_file.write_text(
        """[tool.pysen]
x = 0
[tool.foo]
y = 1"""
    )
    assert pyproject._check_section_exists(test_file)


def test_find_config(temp_dir: pathlib.Path) -> None:
    pysen_config = """[tool.pysen]
x = 0"""
    other_config = """[tool.foo]
x = 0"""

    pyproject_file = temp_dir / "pyproject.toml"
    pysen_file = temp_dir / "pysen.toml"
    other_file = temp_dir / "foo.toml"

    other_file.write_text(pysen_config)
    assert pyproject.find_config(temp_dir) is None

    pyproject_file.write_text(other_config)
    assert pyproject.find_config(temp_dir) is None

    pyproject_file.write_text(pysen_config)
    assert pyproject.find_config(temp_dir) == pyproject_file

    # checks that the config resolution order is pysen_file -> pyproject_file
    pysen_file.write_text(pysen_config)
    assert pyproject.find_config(temp_dir) == pysen_file

    pysen_file.write_text(other_config)
    assert pyproject.find_config(temp_dir) == pyproject_file


def test_find_config_recursive(temp_dir: pathlib.Path) -> None:
    assert pyproject.find_config_recursive(BASE_DIR) == PYPROJECT_PATH
    assert pyproject.find_config_recursive(BASE_DIR / "fakes/configs") == PYPROJECT_PATH
    assert pyproject.find_config_recursive(EXAMPLE_DIR) == EXAMPLE_PYPROJECT_PATH
    assert pyproject.find_config_recursive(temp_dir) is None

    temp_pyproject = temp_dir / "pyproject.toml"
    temp_pyproject.write_text(
        """[tool.pysen]
x = 1"""
    )
    assert pyproject.find_config_recursive(temp_dir) == temp_pyproject



import pathlib
from typing import Optional, Sequence

from pysen import ComponentBase, Manifest, ManifestBase, Source, factory

# raise error while loading
raise BufferError()


def build(  # type: ignore
    components: Sequence[ComponentBase], src_path: Optional[pathlib.Path]
) -> ManifestBase:
    assert src_path is not None
    components = factory.configure_lint(
        factory.ConfigureLintOptions(
            enable_flake8=True,
            enable_isort=True,
            source=Source(includes=[src_path.resolve().parent]),
        )
    )
    return Manifest(components)



import argparse
import pathlib
import tempfile
import unittest.mock
from typing import Optional, Sequence

import pytest

import pysen
from pysen import ConfigureLintOptions, configure_lint
from pysen.exceptions import CommandNotFoundError
from pysen.manifest import Manifest, TargetType
from pysen.reporter import ReporterFactory
from pysen.runner_options import RunOptions

BASE_DIR = pathlib.Path(__file__).resolve().parent


def test_load_manifest() -> None:
    manifest = pysen.load_manifest(BASE_DIR / "fakes/configs/example.toml")
    assert manifest is not None

    with pytest.raises(FileNotFoundError):
        pysen.load_manifest(BASE_DIR / "no_such_file.toml")


def test_build_manifest() -> None:
    # NOTE(igarashi): since build_manifest is just a reference for pysen.build_manifest.build,
    # we just check if the function does not raise an error in this test.
    manifest = pysen.build_manifest(
        [], external_builder=BASE_DIR / "fakes/configs/good_builder.py"
    )
    assert manifest is not None


def test_run() -> None:
    with unittest.mock.patch(
        "pysen.runner.Runner.export_settings"
    ) as mock_export, unittest.mock.patch("pysen.runner.run_target") as mock_run:
        assert pysen.run(
            BASE_DIR, "lint", pyproject=BASE_DIR / "fakes/configs/example.toml"
        )
        mock_export.assert_called()

        # check if settings_dir is correctly handled
        mock_export.reset_mock()
        with tempfile.TemporaryDirectory() as d:
            td = pathlib.Path(d)
            assert pysen.run(
                BASE_DIR,
                "lint",
                pyproject=BASE_DIR / "fakes/configs/example.toml",
                settings_dir=td,
            )
            mock_export.assert_called_once_with(
                BASE_DIR, td, argparse.Namespace(disable=None, enable=None)
            )

        with pytest.raises(CommandNotFoundError):
            assert pysen.run(
                BASE_DIR, "lint2", pyproject=BASE_DIR / "fakes/configs/example.toml"
            )

        components = configure_lint(ConfigureLintOptions(enable_black=True))
        assert pysen.run(BASE_DIR, "lint", components=components)
        with pytest.raises(CommandNotFoundError):
            assert pysen.run(BASE_DIR, "lint2", components=components)

        manifest = Manifest(components)
        assert pysen.run(BASE_DIR, "lint", manifest=manifest)
        with pytest.raises(CommandNotFoundError):
            assert pysen.run(BASE_DIR, "lint2", manifest=manifest)

        # TODO(igarashi): Add test to check run() handles both args and manifest_args

        with pytest.raises(FileNotFoundError):
            pysen.run(BASE_DIR, "lint", pyproject=BASE_DIR / "no_such_file.toml")

        with pytest.raises(ValueError):
            pysen.run(BASE_DIR, "lint")

        with pytest.raises(ValueError):
            pysen.run(
                BASE_DIR,
                "lint",
                pyproject=BASE_DIR / "fakes/configs/example.toml",
                manifest=manifest,
            )

        # NOTE(igarashi): Check that run() returns False when the command reports an error

        def side_effect(
            target: TargetType,
            reporters: ReporterFactory,
            options: RunOptions,
            files: Optional[Sequence[pathlib.Path]],
        ) -> None:
            with reporters.create("hoge") as r:
                r.set_result(False, 128)

        mock_run.side_effect = side_effect

        assert not pysen.run(
            BASE_DIR, "lint", pyproject=BASE_DIR / "fakes/configs/example.toml"
        )
        assert not pysen.run(BASE_DIR, "lint", components=components)
        assert not pysen.run(BASE_DIR, "lint", manifest=manifest)



import argparse
import pathlib
from typing import Dict, List, Mapping, Optional, Sequence

from pysen.command import CommandBase
from pysen.manifest import (
    ComponentName,
    ManifestBase,
    ParserType,
    TargetName,
    TargetType,
)
from pysen.runner_options import PathContext, RunOptions


class FakeManifest(ManifestBase):
    def __init__(
        self,
        expected_base_dir: pathlib.Path,
        expected_settings_dir: Optional[pathlib.Path],
        num_required: bool,
        items: Mapping[str, Sequence[CommandBase]],
        special_item: Sequence[CommandBase],
    ) -> None:
        self._expected_base_dir = expected_base_dir
        self._expected_settings_dir = expected_settings_dir
        self._num_required = num_required
        self._items = {k: list(v) for k, v in items.items()}
        self._special_item = list(special_item)
        # for test assertions
        self._latest_args: Optional[argparse.Namespace] = None

    def configure_parser(self, parser: ParserType) -> None:
        parser.add_argument("--special", action="store_true")
        parser.add_argument("--num", type=int, default=0, required=self._num_required)

    def _targets(self, args: argparse.Namespace) -> Dict[str, TargetType]:
        ret: Dict[str, TargetType] = self._items.copy()
        if args.special:
            ret["special"] = self._special_item

        length = args.num
        if length > 0:
            for k in ret.keys():
                v = ret[k][:length]
                ret[k] = v

        return ret

    def export_settings(self, paths: PathContext, args: argparse.Namespace) -> None:
        assert paths.base_dir == self._expected_base_dir
        if self._expected_settings_dir is not None:
            assert paths.settings_dir == self._expected_settings_dir
        self._latest_args = args

    def get_targets(self, args: argparse.Namespace) -> Dict[str, List[ComponentName]]:
        targets = self._targets(args)
        return {name: [x.name for x in value] for name, value in targets.items()}

    def get_target(
        self,
        target: TargetName,
        paths: PathContext,
        options: RunOptions,
        args: argparse.Namespace,
    ) -> TargetType:
        assert paths.base_dir == self._expected_base_dir
        if self._expected_settings_dir is not None:
            assert paths.settings_dir == self._expected_settings_dir
        return self._targets(args)[target]

    @property
    def latest_args(self) -> Optional[argparse.Namespace]:
        return self._latest_args

    @property
    def clear_latest_args(self) -> None:
        self._latest_args = None



import pathlib

from pysen.path import change_dir, get_relative_path, is_contained, is_covered

BASE_DIR = pathlib.Path(__file__).resolve().parent


def test_get_relative_path() -> None:
    base_dir = pathlib.Path("/opt/pysen/python/packages")
    assert get_relative_path("foo/A", base_dir) == "foo/A"
    assert get_relative_path(pathlib.Path("foo/A"), base_dir) == "foo/A"

    assert get_relative_path("../foo/B", base_dir) == "../foo/B"
    assert get_relative_path(pathlib.Path("../foo/B"), base_dir) == "../foo/B"

    assert get_relative_path("/opt/pysen/C", base_dir) == "../../C"
    assert get_relative_path(pathlib.Path("/opt/pysen/C"), base_dir) == "../../C"

    assert get_relative_path("/opt/pysen/python2/D", base_dir) == "../../python2/D"
    assert (
        get_relative_path("/opt/pysen/python/packages/configs/E", base_dir)
        == "configs/E"
    )
    assert (
        get_relative_path("/home/user/.config/F", base_dir)
        == "../../../../home/user/.config/F"
    )


def test_change_dir() -> None:
    current = pathlib.Path.cwd()

    with change_dir(BASE_DIR / "fakes"):
        assert pathlib.Path.cwd() == BASE_DIR / "fakes"

    assert pathlib.Path.cwd() == current


def test_is_covered() -> None:
    hoge = pathlib.Path("hoge")
    bar = pathlib.Path("bar")
    assert is_covered(hoge, [hoge])
    assert is_covered(hoge, [hoge, bar])
    assert is_covered(hoge / "foo", [hoge])
    assert is_covered(hoge / "foo" / "bar", [hoge])

    assert not is_covered(hoge, [])
    assert not is_covered(hoge, [bar])
    assert not is_covered(hoge, [hoge / "subdir"])


def test_is_contained() -> None:
    foo = pathlib.Path("foo").resolve()
    bar = pathlib.Path("bar")
    baz = pathlib.Path("baz")
    ufoo = pathlib.Path("~/foo").resolve()
    ufooo = pathlib.Path("~/foo").expanduser().resolve()

    assert not is_contained(foo / bar, foo)
    assert not is_contained(foo / bar, foo / baz / bar)
    assert is_contained(foo, foo / bar)
    assert is_contained(foo / bar, foo / bar / baz)

    # user expansion not supported
    assert not is_contained(ufoo, ufooo)



from typing import Any, Callable, Optional, Sequence, Set, Tuple


def foo(
    a: Any,
    b: Callable[[], Tuple[int, int, str]],
    c: Set[str],
    d: Optional[Sequence[int]] = None,
    e: Any = None,
) -> None:
    pass


print("Hello world")
foo(a=1, b=lambda: (1, 2, "hoge"), c=set(), d=None, e=None)



import pathlib
from unittest import mock

import pytest

from pysen.diagnostic import Diagnostic, FLCMFormatter
from pysen.reporter import Reporter, ReporterFactory

BASE_DIR = pathlib.Path(__file__).resolve().parent


def test_reporter() -> None:
    r = Reporter("fuga")

    with pytest.raises(AssertionError):
        r.success

    with pytest.raises(AssertionError):
        r.elapsed_time

    with mock.patch("time.time") as m:
        m.return_value = 100.0
        with r:
            assert r.name == "fuga"
            r.report_command("command --flag")
            r.set_result(True, 0)

            m.return_value = 123.4

    assert r.elapsed_time == pytest.approx(23.4)
    assert r.success

    r = Reporter("piyo")
    with r:
        assert r.name == "piyo"
        r.set_result(False, 128)

    assert not r.success


def test_reporter_report_diagnostics() -> None:
    r = Reporter("foo")
    d1 = Diagnostic(pathlib.Path("hoge").resolve(), 1, 2, 3, message="hoge")
    d2 = Diagnostic(pathlib.Path("fuga").resolve(), 4, 5, 6, message="fuga")
    d3 = Diagnostic(pathlib.Path("piyo").resolve(), 7, 8, 9, message="piyo")

    with r:
        r.report_diagnostics([d1])
        assert r.diagnostics == [d1]

        r.report_diagnostics([d2, d3])
        assert r.diagnostics == [d1, d2, d3]


def test_reporter_factory() -> None:
    factory = ReporterFactory()
    assert len(factory.reporters) == 0

    with factory.create("foo") as r:
        r.set_result(True, 0)
        r.report_diagnostics(
            [Diagnostic(BASE_DIR / "hoge.py", 1, 2, 3, message="error")]
        )

    assert len(factory.reporters) == 1
    assert not factory.has_error()
    out = factory.format_summary()
    assert "foo" in out

    with factory.create("bar") as r:
        r.set_result(False, 128)

    assert len(factory.reporters) == 2
    assert factory.has_error()
    out = factory.format_summary()
    assert "foo" in out and "bar" in out

    err_summary = factory.format_error_summary()
    assert "\n - bar\n" in err_summary and "foo" not in err_summary

    out = factory.format_diagnostic_summary(FLCMFormatter)
    assert f"{BASE_DIR / 'hoge.py'}:1:3:foo: error" in out



import pathlib

from pysen.cli_config import parse

BASE_DIR = pathlib.Path(__file__).resolve().parent
CONFIG_DIR = BASE_DIR / "fakes/configs"


def test_example() -> None:
    config = parse(CONFIG_DIR / "example.toml")
    assert config is not None
    assert config.settings_dir is not None
    assert config.settings_dir == CONFIG_DIR / "hoge"
    assert config.settings_dir.is_absolute()


def test_parse() -> None:
    assert parse(CONFIG_DIR / "example.toml") is not None
    assert parse(CONFIG_DIR / "simple_source.toml") is None



import io
import logging
import os
import pathlib
import sys
import tempfile
from typing import List

import pytest

from pysen.process_utils import _read_stream, run
from pysen.reporter import Reporter

SAMPLE_DATA = """BytesIO example string.

Motivation-Driven
Learn or Die
Proud, but Humble
Boldly do what no one has done before"""

SAMPLE_SCRIPT = """#!/bin/bash
# This string is used to check if trailing new lines are kept.

echo "Start"
echo "StartError" > /dev/stderr

# Test sleep
sleep 0.1

for x in {0..99}
do
  echo "out${x}"
  echo "err${x}" > /dev/stderr
done

echo "End"
echo "EndError" > /dev/stderr

"""

TEST_UNICODE = b"pysen\xe3\x81\xb1\xe3\x81\x84\xe3\x81\x9b\xe3\x82\x9312345"


class FakeHandler(logging.Handler):
    def __init__(self) -> None:
        super().__init__()
        self._messages: List[str] = []

    @property
    def messages(self) -> List[str]:
        return self._messages

    def emit(self, record: logging.LogRecord) -> None:
        self._messages.append(record.msg)


class HandlerException(Exception):
    pass


class FailingHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        raise HandlerException("somebody screwed up")


@pytest.mark.parametrize(
    "sample_str",
    [SAMPLE_DATA, SAMPLE_SCRIPT],
    ids=["sample", "sample_with_trailing_newlines"],
)
def test__read_stream(sample_str: str) -> None:
    # NOTE(igarashi): Since BytesIO does not inherit RawIOBase but BufferedIOBase,
    # we cannot instantiate BufferedReader from bytes directly.
    # We use FileIOBase instead.
    with tempfile.TemporaryDirectory() as td:
        temp_dir = pathlib.Path(td)
        temp_path = temp_dir / "file"
        temp_path.write_text(sample_str)

        reporter = Reporter("foo")
        handler = FakeHandler()
        reporter.process_output.setLevel(logging.INFO)
        reporter.process_output.handlers.clear()
        reporter.process_output.addHandler(handler)

        ret = _read_stream(io.StringIO(temp_path.read_text()), reporter, logging.INFO)
        expected = sample_str
        assert ret == expected
        assert handler.messages == expected.splitlines()

        handler.messages.clear()
        ret = _read_stream(io.StringIO(temp_path.read_text()), reporter, logging.DEBUG)

        assert ret == expected
        assert handler.messages == []


def test_run_encoding() -> None:
    with tempfile.TemporaryDirectory() as td:
        temp_dir = pathlib.Path(td)
        temp_path = temp_dir / "file"
        temp_path.write_bytes(TEST_UNICODE)
        reporter = Reporter("foo")

        expected_stdout = temp_path.read_text(sys.stdout.encoding, errors="ignore")
        expected_utf8 = temp_path.read_text("utf-8", errors="ignore")
        expected_ascii = temp_path.read_text("ascii", errors="ignore")

        ret, stdout, stderr = run(["cat", str(temp_path)], reporter)
        assert ret == 0
        assert stdout == expected_stdout

        ret, stdout, _ = run(["cat", str(temp_path)], reporter, encoding="utf-8")
        assert ret == 0
        assert stdout == expected_utf8

        ret, stdout, _ = run(["cat", str(temp_path)], reporter, encoding="ascii")
        assert ret == 0
        assert stdout == expected_ascii == "pysen12345"


def test_run() -> None:
    assert os.getenv("LANG", "C") == "C", "Did you run pytest through tox?"
    with tempfile.TemporaryDirectory() as td:
        temp_dir = pathlib.Path(td)
        temp_path = temp_dir / "file"
        temp_path.touch()

        reporter = Reporter("foo")
        handler = FakeHandler()
        reporter.process_output.setLevel(logging.INFO)
        reporter.process_output.handlers.clear()
        reporter.process_output.addHandler(handler)

        ret, stdout, stderr = run(["ls", str(temp_path)], reporter)
        assert ret == 0
        assert "file" in stdout
        assert stderr == ""
        assert len(handler.messages) > 0

        handler.messages.clear()
        ret, stdout, stderr = run(
            ["ls", str(temp_path)], reporter, stdout_loglevel=logging.NOTSET
        )
        assert ret == 0
        assert "file" in stdout
        assert stderr == ""
        assert len(handler.messages) == 0

        handler.messages.clear()
        ret, stdout, stderr = run(["ls", str(temp_dir / "invalid")], reporter)
        assert ret != 0
        assert stdout == ""
        assert "No such file or directory" in stderr
        assert len(handler.messages) > 0

        handler.messages.clear()
        ret, stdout, stderr = run(
            ["ls", str(temp_dir / "invalid")], reporter, stderr_loglevel=logging.DEBUG
        )
        assert ret != 0
        assert stdout == ""
        assert "No such file or directory" in stderr
        assert len(handler.messages) == 0

        # check if run method can handle large output and errors
        temp_path.write_text(SAMPLE_SCRIPT)

        ret, stdout, stderr = run(["bash", str(temp_path)], reporter)

        stdout_lines = stdout.splitlines()
        stderr_lines = stderr.splitlines()

        assert len(stdout_lines) == 102
        assert len(stderr_lines) == 102

        assert stdout_lines[0] == "Start" and stderr_lines[0] == "StartError"
        assert stdout_lines[-1] == "End" and stderr_lines[-1] == "EndError"

        for x in range(100):
            assert stdout_lines[x + 1] == f"out{x}"
            assert stderr_lines[x + 1] == f"err{x}"

        # exceptions encountered in sub-threads shall be raised
        reporter.process_output.addHandler(FailingHandler())
        with pytest.raises(HandlerException):
            ret, stdout, stderr = run(["echo", "mashimashi"], reporter)



from pysen._version import __version__
from pysen.py_version import VersionRepresentation


def test__version() -> None:
    # version string MUST be in a format the VersionRepresentation understands
    VersionRepresentation.from_str(__version__)



from pathlib import Path

from pysen.ext.black_wrapper import _parse_file_path


def test__parse_file_path() -> None:
    black_format = (
        "path/test_error_line_parser.py      2020-06-01 07:19:58.515112 +0000"
    )
    assert _parse_file_path(black_format) == Path("path/test_error_line_parser.py")



from pathlib import Path
from unittest import mock

from pysen.error_lines import parse_error_diffs, parse_error_lines
from pysen.ext.black_wrapper import _parse_file_path

std_err1 = "/path/to/file1.py:70:5: error: Missing return statement [return]\n"
std_err2 = "/path/to/file2.py:71:6: error: Missing return statement [return]\n"
std_err = "".join([std_err1, std_err2])
mypy_invalid_format = (
    "/home/user/pysen/__init__.py:72: error: unused 'type: ignore' comment\n"
)


diff_err1 = """--- /tmp/tmp.py      2020-05-29 13:45:10.907383 +0000
+++ /tmp/tmp.py      2020-05-29 13:45:12.563367 +0000
@@ -1,7 +1,8 @@
 class Hoge:
-    a=3
+    a = 3
+
 
 for a in range(3):
     print(a)
 
 for b in range(3):
@@ -17,6 +18,7 @@
     print(a)
 
 for b in range(3):
     print(b)
 
-answer_to_everything=42
+answer_to_everything = 42
+
"""  # NOQA

diff_err2 = """--- /home/user/pysen/pysen/cli.py  2020-06-10 05:31:01.167304 +0000
+++ /home/user/pysen/pysen/cli.py  2020-06-10 05:35:33.973467 +0000
@@ -147,10 +147,11 @@
     )
     parser.add_argument(
         "--version", action="store_true", help="Show pysen version and exit",
     )
     import math
+
     parser.add_argument(
         "--loglevel", type=str, help="Set loglevel", choices=list(LogLevel.__members__),
     )
     return parser

"""  # NOQA

diff_err3 = """--- /home/user/pysen/pysen/cli.py  2020-06-10 05:31:01.167304 +0000
+++ /home/user/pysen/pysen/cli.py  2020-06-10 05:35:33.973467 +0000
@@ -147,10 +147,9 @@
     )
     parser.add_argument(
         "--version", action="store_true", help="Show pysen version and exit",
     )
     import math
-
     parser.add_argument(
         "--loglevel", type=str, help="Set loglevel", choices=list(LogLevel.__members__),
     )
     return parser

"""  # NOQA

diff_err4 = """--- /home/user/pysen/foo.py        2021-03-25 16:43:53.875256 +0000
+++ /home/user/pysen/foo.py        2021-03-25 16:44:02.267033 +0000
@@ -1,8 +1,5 @@
 {
     foo: "",
-
-
     bar: [],
-
     baz: {},
 }
"""  # NOQA


def test_standard_parser() -> None:
    err1, err2 = parse_error_lines(std_err)

    assert err1.file_path == Path("/path/to/file1.py")
    assert err1.start_line == 70
    assert err1.end_line == 70
    assert err1.start_column == 5
    assert err1.message == "error: Missing return statement [return]"

    assert err2.file_path == Path("/path/to/file2.py")
    assert err2.start_line == 71
    assert err2.end_line == 71
    assert err2.start_column == 6
    assert err2.message == "error: Missing return statement [return]"

    err3 = list(parse_error_lines(mypy_invalid_format))[0]

    assert err3.file_path == Path("/home/user/pysen/__init__.py")
    assert err3.start_line == 72
    assert err3.end_line == 72
    assert err3.start_column is None
    assert err3.message == "error: unused 'type: ignore' comment"

    with mock.patch("pysen.error_lines._warn_parse_error") as warn:
        error = "invalid_format\n"
        list(parse_error_lines(error))
        warn.assert_called_with(error.rstrip("\n"), None)

    logger = mock.Mock()
    with mock.patch("pysen.error_lines._warn_parse_error") as warn:
        error = "invalid_format\n"
        list(parse_error_lines(error, logger))
        warn.assert_called_with(error.rstrip("\n"), logger)


def test_diff_parser() -> None:
    err1, err2 = parse_error_diffs(diff_err1, _parse_file_path)

    assert err1.start_line == 2
    assert err1.end_line == 2
    assert err1.start_column == 1
    assert err1.file_path == Path("/tmp/tmp.py")
    assert err1.diff == "-    a=3\n+    a = 3\n+\n"

    assert err2.start_line == 22
    assert err2.end_line == 22
    assert err2.start_column == 1
    assert err2.file_path == Path("/tmp/tmp.py")
    assert err2.diff == "-answer_to_everything=42\n+answer_to_everything = 42\n+\n"
    with mock.patch("pysen.error_lines._warn_parse_error") as warn:
        diff = "---/tmp/tmp.py\n+++/tmp/tmp.py\n@@ -1,7 +1,8 @@\ninvalid\n"
        list(parse_error_diffs(diff, _parse_file_path))
        warn.assert_called_with(diff, None)

    logger = mock.Mock()
    with mock.patch("pysen.error_lines._warn_parse_error") as warn:
        diff = "---/tmp/tmp.py\n+++/tmp/tmp.py\n@@ -1,7 +1,8 @@\ninvalid\n"
        list(parse_error_diffs(diff, _parse_file_path, logger))
        warn.assert_called_with(diff, logger)

    # has only target diff
    errors = list(parse_error_diffs(diff_err2, _parse_file_path))
    assert len(errors) == 1
    err = errors[0]
    assert err.start_line == 152
    assert err.end_line == 152
    assert err.diff == "+\n     parser.add_argument(\n"

    # has only source diff
    errors = list(parse_error_diffs(diff_err3, _parse_file_path))
    assert len(errors) == 1
    err = errors[0]
    assert err.start_line == 152
    assert err.end_line == 152
    assert err.diff == "-\n\n"

    # has multiple deletion
    errors = list(parse_error_diffs(diff_err4, _parse_file_path))

    assert len(errors) == 1
    err1 = errors[0]
    assert err1.start_line == 3
    assert err1.end_line == 6
    assert err1.diff == "-\n-\n     bar: [],\n-\n"



import logging
import pathlib
from typing import Any, List

import dacite
import pytest
import tomlkit
from _pytest.logging import LogCaptureFixture

from pysen.factory import MypyModuleOption
from pysen.isort import IsortSectionName
from pysen.mypy import MypyFollowImports, MypyPreset
from pysen.py_version import PythonVersion, VersionRepresentation
from pysen.pyproject_model import (
    Config,
    InvalidConfigurationError,
    LintConfig,
    _load_version,
    _migrate_alias_fields,
    _migrate_deprecated_fields,
    _parse_mypy_modules,
    _parse_mypy_target,
    _parse_mypy_targets,
    _parse_plugin_configs,
    _parse_python_version,
    _parse_source,
    has_tool_section,
    parse,
)
from pysen.source import SourceEntrySetting

BASE_DIR = pathlib.Path(__file__).resolve().parent
CONFIG_DIR = BASE_DIR / "fakes/configs"


def test__parse_plugin_configs() -> None:
    data_x = {
        "script": "hoge.py",
    }
    data_y = {
        "function": "foo.bar::baz",
        "config": {"message": "hello"},
    }
    data = {
        "X": data_x,
        "Y": data_y,
    }

    configs = _parse_plugin_configs(BASE_DIR, data)

    assert len(configs) == 2
    assert configs[0].location == "tool.pysen.plugin.X"
    assert configs[1].location == "tool.pysen.plugin.Y"
    assert configs[0].script == BASE_DIR / "hoge.py"
    assert configs[1].script is None
    assert configs[0].function is None
    assert configs[1].function == "foo.bar::baz"

    assert configs[0].config is None
    assert configs[1].config == {
        "message": "hello",
    }

    data_invalid_path = {
        "location": "don't set this section from pyproject",
        "script": "hoge.py",
    }
    with pytest.raises(dacite.DaciteError) as ex:
        _parse_plugin_configs(BASE_DIR, {"X": data_invalid_path})
    assert "unknown filed" in str(ex.value)

    data_invalid_no_plugin = {
        "config": {"message": "hello"},
    }
    with pytest.raises(dacite.DaciteError) as ex:
        _parse_plugin_configs(BASE_DIR, {"X": data_invalid_no_plugin})
    assert "must specify" in str(ex.value)

    data_invalid_both = {
        "function": "foo.bar::baz",
        "script": "hoge.py",
        "config": {"message": "hello"},
    }
    with pytest.raises(dacite.DaciteError) as ex:
        _parse_plugin_configs(BASE_DIR, {"X": data_invalid_both})
    assert "only one of" in str(ex.value)

    with pytest.raises(dacite.WrongTypeError):
        _parse_plugin_configs(BASE_DIR, [])


def test__parse_source() -> None:
    source = _parse_source(BASE_DIR, ["foo", "bar/baz"])
    assert len(source.includes) == 2
    assert source.includes.keys() == {BASE_DIR / "foo", BASE_DIR / "bar/baz"}
    assert all(not x.glob for x in source.includes.values())

    source = _parse_source(
        BASE_DIR,
        {
            "includes": ["foo"],
            "include_globs": ["bar/*.template"],
            "excludes": ["hoge"],
            "exclude_globs": ["fuga/*_grpc.py"],
        },
    )
    includes = source.includes
    assert len(includes) == 2
    assert includes.keys() == {BASE_DIR / "foo", "bar/*.template"}
    assert not includes[BASE_DIR / "foo"].glob
    assert includes["bar/*.template"].glob
    assert includes["bar/*.template"].base_dir == BASE_DIR

    excludes = source.excludes
    assert len(excludes) == 2
    assert excludes.keys() == {BASE_DIR / "hoge", "fuga/*_grpc.py"}
    assert not excludes[BASE_DIR / "hoge"].glob
    assert excludes["fuga/*_grpc.py"].glob
    assert excludes["fuga/*_grpc.py"].base_dir == BASE_DIR

    source = _parse_source(
        BASE_DIR,
        {"excludes": ["hoge"], "exclude_globs": ["fuga/*_grpc.py"]},
    )

    includes = source.includes
    assert len(includes) == 1
    assert includes.keys() == {"."}
    assert not includes["."].glob
    assert includes["."].base_dir == BASE_DIR

    excludes = source.excludes
    assert len(excludes) == 2
    assert excludes.keys() == {BASE_DIR / "hoge", "fuga/*_grpc.py"}

    with pytest.raises(dacite.DaciteError):
        _parse_source(BASE_DIR, 1.0)


def test__parse_python_version() -> None:
    assert _parse_python_version("py38") == PythonVersion(3, 8)
    assert _parse_python_version("PY38") == PythonVersion(3, 8)
    with pytest.raises(dacite.DaciteError) as ex:
        _parse_python_version("PY999")

    assert "one of" in str(ex.value)  # ensure that we suggest some options

    with pytest.raises(dacite.WrongTypeError):
        _parse_python_version(38)


def test__parse_mypy_target() -> None:
    base_dir = pathlib.Path("/foo")

    with pytest.raises(dacite.DaciteError):
        _parse_mypy_target(base_dir, "a")

    with pytest.raises(dacite.DaciteError):
        _parse_mypy_target(base_dir, {"paths": []})

    target = _parse_mypy_target(base_dir, {"paths": ["a", "b", "/d"]})
    assert target.paths == [base_dir / "a", base_dir / "b", pathlib.Path("/d")]

    target = _parse_mypy_target(base_dir, {"paths": ["a"], "namespace_packages": True})
    assert target.paths == [base_dir / "a"]
    assert target.namespace_packages


def test__parse_mypy_targets() -> None:
    base_dir = pathlib.Path("/foo")

    with pytest.raises(dacite.DaciteError) as e:
        _parse_mypy_targets(base_dir, ["a", "b", "/c"])

    base_dir = pathlib.Path("/foo")
    test_data: List[Any] = [{"paths": ["x", "/y", "z"]}, {"paths": ["a", "b"]}]
    targets = _parse_mypy_targets(base_dir, test_data)
    assert len(targets) == 2
    assert targets[0].paths == [base_dir / "x", pathlib.Path("/y"), base_dir / "z"]
    assert targets[1].paths == [base_dir / "a", base_dir / "b"]

    test_data.append("x")
    with pytest.raises(dacite.DaciteError) as e:
        _parse_mypy_targets(base_dir, test_data)

    assert "tool.pysen.lint.mypy_targets must be a list of dicts" in str(e.value)


def test__parse_mypy_modules() -> None:
    with pytest.raises(dacite.WrongTypeError):
        _parse_mypy_modules("x")

    assert _parse_mypy_modules({}) == {}

    modules = _parse_mypy_modules({"a": {}, "b": {}})
    assert modules.keys() == {"a", "b"}
    assert not modules["a"].ignore_errors
    assert not modules["b"].ignore_errors
    assert modules["a"].preset is None
    assert modules["b"].preset is None

    with pytest.raises(dacite.WrongTypeError):
        _parse_mypy_modules({"a": "b"})

    with pytest.raises(dacite.WrongTypeError):
        _parse_mypy_modules({1: {}})

    modules = _parse_mypy_modules(
        {"a": {"preset": "entry"}, "b.c": {"ignore_errors": True}}
    )
    assert modules.keys() == {"a", "b.c"}
    assert not modules["a"].ignore_errors
    assert modules["b.c"].ignore_errors
    assert modules["a"].preset == MypyPreset.ENTRY
    assert modules["b.c"].preset is None

    with pytest.raises(dacite.DaciteError) as e:
        _parse_mypy_modules({"a.b.c": {"preset": "entry", "ignore_errors": True}})

    assert "a.b.c" in str(e.value)


def test__migrate_alias_fields() -> None:
    config = Config(lint=None)
    _migrate_alias_fields(config)
    assert config.lint is None

    config = Config(lint=LintConfig(mypy_ignore_packages=["X", "Y.Z.*"]))
    _migrate_alias_fields(config)
    assert config.lint is not None
    assert config.lint.mypy_ignore_packages is None
    assert config.lint.mypy_modules is not None
    assert config.lint.mypy_modules.keys() == {"X", "Y.Z.*"}
    assert config.lint.mypy_modules["X"].ignore_errors
    assert config.lint.mypy_modules["Y.Z.*"].ignore_errors

    config = Config(
        lint=LintConfig(
            mypy_ignore_packages=["X"], mypy_modules={"X": MypyModuleOption()}
        )
    )
    with pytest.raises(dacite.DaciteError) as e:
        _migrate_alias_fields(config)

    assert "X is configured in both mypy_ignore_packages and mypy_modules" in str(
        e.value
    )


def test__migrate_deprecated_fields(caplog: LogCaptureFixture) -> None:
    base_dir = pathlib.Path("/foo")
    setting_path = pathlib.Path("/foo/pyproject.toml")

    caplog.clear()
    config = Config(lint=None)
    _migrate_deprecated_fields(setting_path, config)
    assert config.lint is None
    assert caplog.records == []

    config = Config(
        lint=LintConfig(
            mypy_target_dirs=[base_dir / "a", base_dir / "b", base_dir / "c"]
        ),
    )
    _migrate_deprecated_fields(setting_path, config)
    assert config.lint is not None
    assert config.lint.mypy_target_dirs is None
    assert config.lint.mypy_targets is not None and len(config.lint.mypy_targets) == 1
    assert config.lint.mypy_targets[0].paths == [
        base_dir / "a",
        base_dir / "b",
        base_dir / "c",
    ]
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelno == logging.WARNING
    assert "tool.pysen.mypy_target_dirs is deprecated" in record.message
    assert str(setting_path) in record.message


def test_example() -> None:
    config = parse(CONFIG_DIR / "example.toml")
    assert config is not None

    assert config.builder is None
    assert config.lint is not None
    lint = config.lint
    assert lint.enable_black
    assert not lint.enable_flake8
    assert lint.enable_isort
    assert lint.enable_mypy
    assert lint.line_length == 88
    assert isinstance(lint.line_length, int)
    assert lint.py_version == PythonVersion(3, 8)
    assert lint.isort_known_first_party == ["alpha"]
    assert lint.isort_known_third_party == ["beta", "gamma"]
    assert lint.isort_default_section == IsortSectionName.THIRDPARTY
    assert lint.mypy_modules is not None
    assert lint.mypy_modules.keys() == {"pysen.stubs", "pysen.proto", "apple", "banana"}
    assert lint.mypy_modules["pysen.stubs"].ignore_errors
    assert lint.mypy_modules["pysen.proto"].ignore_errors
    assert not lint.mypy_modules["apple"].ignore_errors
    assert lint.mypy_modules["banana"].ignore_errors

    assert lint.mypy_modules["pysen.stubs"].follow_imports == MypyFollowImports.SKIP
    assert lint.mypy_modules["pysen.proto"].follow_imports == MypyFollowImports.SKIP
    assert lint.mypy_modules["apple"].follow_imports == MypyFollowImports.SILENT
    assert lint.mypy_modules["banana"].follow_imports is None

    assert lint.mypy_modules["apple"].preset == MypyPreset.ENTRY
    assert lint.mypy_path == [CONFIG_DIR / "pysen-stubs"]

    assert lint.mypy_plugins is not None
    assert len(lint.mypy_plugins) == 2
    assert lint.mypy_plugins[0].script == CONFIG_DIR / pathlib.Path("./sugoi/plugin")
    assert lint.mypy_plugins[1].function == "sugoi_plugin:entry"

    assert lint.source is not None
    source = lint.source
    assert source.includes == {
        CONFIG_DIR: SourceEntrySetting(glob=False),
        CONFIG_DIR / "hoge": SourceEntrySetting(glob=False),
        "**/*.template": SourceEntrySetting(glob=True, base_dir=CONFIG_DIR),
    }
    assert source.excludes == {
        CONFIG_DIR / "fuga": SourceEntrySetting(glob=False),
        "foo/*_pb2.py": SourceEntrySetting(glob=True, base_dir=CONFIG_DIR),
    }


def test_simple_source() -> None:
    config = parse(CONFIG_DIR / "simple_source.toml")
    assert config is not None

    assert config.builder is None
    assert config.lint is not None
    lint = config.lint
    assert lint.enable_mypy
    assert lint.line_length == 80
    assert lint.py_version == PythonVersion(2, 7)
    assert lint.source is not None
    source = lint.source
    assert source.includes == {
        CONFIG_DIR: SourceEntrySetting(glob=False),
        CONFIG_DIR / "hoge": SourceEntrySetting(glob=False),
        CONFIG_DIR / "piyo": SourceEntrySetting(glob=False),
    }
    assert source.excludes == {}


def test_builder() -> None:
    config = parse(CONFIG_DIR / "builder.toml")
    assert config is not None

    assert config.lint is None
    assert config.builder is not None
    # check if builder is already resolved (is an abspath)
    assert config.builder.is_absolute()
    assert config.builder == CONFIG_DIR / "good_builder.py"


def test_plugin() -> None:
    config = parse(CONFIG_DIR / "plugin.toml")
    assert config is not None

    assert config.lint is None
    assert config.plugin is not None
    assert len(config.plugin) == 2
    assert config.plugin[0].location == "tool.pysen.plugin.hoge"
    assert config.plugin[0].script == CONFIG_DIR / "../plugin.py"
    assert config.plugin[0].function is None
    assert config.plugin[0].config is None

    assert config.plugin[1].location == "tool.pysen.plugin.fuga"
    assert config.plugin[1].script is None
    assert config.plugin[1].function == "fakes.plugins::create"
    assert config.plugin[1].config == {
        "message": "hello",
        "value": 10.0,
        "flag": False,
    }


def test_lint_config_update() -> None:
    lhs = LintConfig(
        base=pathlib.Path("hoge"),
        enable_black=True,
        enable_isort=True,
        line_length=80,
        isort_known_first_party=["hoge", "fuga"],
        isort_known_third_party=["piyo"],
    )
    rhs = LintConfig(
        base=pathlib.Path("fuga"),
        enable_black=False,
        enable_flake8=True,
        isort_known_first_party=["foo"],
        py_version=PythonVersion(3, 8),
    )

    lhs.update(rhs)

    assert lhs.base == pathlib.Path("hoge")
    assert not lhs.enable_black
    assert lhs.enable_flake8
    assert lhs.enable_isort
    assert not lhs.enable_mypy
    assert lhs.line_length == 80
    assert lhs.isort_known_first_party == ["foo"]
    assert lhs.isort_known_third_party == ["piyo"]
    assert lhs.py_version == PythonVersion(3, 8)


def test_has_tool_section() -> None:
    def _check_example_toml(tool_name: str, filename: str) -> bool:
        path = CONFIG_DIR / filename
        pyproject = tomlkit.loads(path.read_text())
        return has_tool_section(tool_name, pyproject)

    assert not _check_example_toml("pysen", "non_pysen_config.toml")
    assert not _check_example_toml("jiro", "non_pysen_config.toml")
    assert _check_example_toml("pysen", "base.toml")
    assert _check_example_toml("pysen", "base2.toml")
    assert _check_example_toml("pysen", "empty.toml")
    assert _check_example_toml("pysen", "example.toml")
    assert _check_example_toml("pysen", "builder.toml")
    assert _check_example_toml("pysen", "simple_source.toml")


def test___load_version() -> None:
    assert _load_version({}) is None
    assert _load_version({"version": "0.1.2a1"}) == VersionRepresentation(0, 1, 2, "a1")
    with pytest.raises(InvalidConfigurationError):
        _load_version({"version": "none"})



from pysen.black import Black
from pysen.factory import ConfigureLintOptions, configure_lint
from pysen.source import Source


def test_configure_lint_default_source() -> None:
    components = configure_lint(ConfigureLintOptions(enable_black=True, source=None))
    assert len(components) == 1
    black = components[0]
    assert isinstance(black, Black)
    source = black.source
    assert isinstance(source, Source)
    assert source.includes.keys() == {"."}



import pathlib
import sys

import pytest

from pysen.py_module import _parse_entry_point, load

CURRENT_FILE = pathlib.Path(__file__).resolve()
BASE_DIR = CURRENT_FILE.parent


def test__parse_entry_point() -> None:
    assert _parse_entry_point("hoge") is None
    assert _parse_entry_point("hoge.fuga") is None
    assert _parse_entry_point("hoge.fuga::piyo") == ("hoge.fuga", "piyo")
    assert _parse_entry_point("hoge::piyo") == ("hoge", "piyo")
    assert _parse_entry_point("::piyo") is None
    assert _parse_entry_point("1::piyo") is None
    assert _parse_entry_point("hoge::2") is None
    assert _parse_entry_point("_hoge1._fuga2_::_piyo3_") == (
        "_hoge1._fuga2_",
        "_piyo3_",
    )
    assert _parse_entry_point("hoge::piyo\nfuga") is None


def test_load() -> None:
    with pytest.raises(FileNotFoundError):
        load(BASE_DIR / "hoge", "builder")

    with pytest.raises(FileNotFoundError):
        load(BASE_DIR, "builder")

    module = load(BASE_DIR / "fakes/configs/good_builder.py", "foo")
    assert module is not None
    assert getattr(module, "build") is not None  # NOQA: B009
    assert module.__name__.startswith("pysen._modules.foo_")
    assert sys.modules[module.__name__]

    module = load(BASE_DIR / "fakes/configs/invalid_interface_builder.py", "bar")
    assert module is not None
    assert getattr(module, "build2") is not None  # NOQA: B009
    assert module.__name__.startswith("pysen._modules.bar_")
    assert sys.modules[module.__name__]

    module2 = load(BASE_DIR / "fakes/configs/invalid_return_builder.py", "foo")
    assert module2 is not None
    assert getattr(module2, "build") is not None  # NOQA: B009
    assert sys.modules[module2.__name__]
    assert module2.__name__ != module.__name__

    # NOTE(igarashi): py_module.load currently raises the error that a loading module
    # raises as is. It might be better to wrap the exception with our custom error type
    # so that the one can know the error is from a module.
    with pytest.raises(BufferError):
        load(BASE_DIR / "fakes/configs/error_builder.py", "builder")



import dataclasses
import pathlib
from typing import List, Sequence

import dacite

from fakes.component import FakeComponent, Operation
from pysen.component import ComponentBase
from pysen.plugin import PluginBase
from pysen.pyproject_model import Config, PluginConfig


@dataclasses.dataclass
class FakePluginConfig:
    enable_c1: bool = False
    enable_c2: bool = False


class FakePlugin(PluginBase):
    def __init__(self, name: str) -> None:
        self._name = name

    def load(
        self, file_path: pathlib.Path, config_data: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        config = FakePluginConfig()
        if config_data.config is not None:
            config = dacite.from_dict(
                FakePluginConfig, config_data.config, dacite.Config(strict=True)
            )

        r = [0.0]
        components: List[ComponentBase] = []

        if config.enable_c1:
            components.append(
                FakeComponent(
                    f"{self._name}_component1",
                    {"op1": (2, Operation.MUL), "op2": (10, Operation.ADD)},
                    None,
                    None,
                    r,
                )
            )

        if config.enable_c2:
            components.append(
                FakeComponent(
                    f"{self._name}_component2",
                    {"op1": (3, Operation.MUL), "op3": (-1, Operation.MUL)},
                    None,
                    None,
                    r,
                )
            )

        return components


def create() -> PluginBase:
    return FakePlugin("create")


def plugin() -> PluginBase:
    return FakePlugin("plugin")



import logging
import pathlib
from tempfile import TemporaryDirectory
from typing import List, Optional, Set
from unittest import mock

from pysen.diagnostic import Diagnostic
from pysen.lint_command import (
    LintCommandBase,
    SingleFileFormatCommandBase,
    SingleFileLintCommandBase,
)
from pysen.reporter import Reporter
from pysen.source import FilePredicateType, Source

BASE_DIR = pathlib.Path(__file__).resolve().parent


class FakeLintCommand(LintCommandBase):
    @property
    def name(self) -> str:
        return "fake"

    def __call__(self, reporter: Reporter) -> int:
        pass


class FakeHandler(logging.Handler):
    def __init__(self) -> None:
        super().__init__()
        self._messages: List[str] = []

    @property
    def messages(self) -> List[str]:
        return self._messages

    def emit(self, record: logging.LogRecord) -> None:
        self._messages.append(record.msg)


class FakeSource(Source):
    def resolve_files(
        self,
        base_dir: pathlib.Path,
        filter_predicate: Optional[FilePredicateType],
        use_git: bool = True,
        reporter: Optional[Reporter] = None,
    ) -> Set[pathlib.Path]:
        assert filter_predicate is not None

        files = filter(
            filter_predicate, map(pathlib.Path, ["foo.py", "bar.pyi", "baz.txt"])
        )
        return {base_dir / f for f in files}


class FakeSingleFileLintCommand(SingleFileLintCommandBase):
    @property
    def name(self) -> str:
        return "fake"

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.suffix in {".py", ".pyi"}

    def check(self, file_path: pathlib.Path, reporter: Reporter) -> bool:
        pass


class FakeSingleFileFormatCommand(SingleFileFormatCommandBase):
    @property
    def name(self) -> str:
        return "fake"

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.suffix in {".py", ".pyi"}

    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        pass


def test_lint_command_base() -> None:
    source = Source()
    command = FakeLintCommand(BASE_DIR, source)

    assert command.name == "fake"
    assert command.base_dir == BASE_DIR
    assert command.source == source

    with mock.patch("pysen.git_utils.check_git_available", return_value=True):
        assert command.git_enabled()

    with mock.patch("pysen.git_utils.check_git_available", return_value=False):
        assert not command.git_enabled()


def test_single_file_lint_command_base() -> None:
    with TemporaryDirectory() as t:
        base_dir = pathlib.Path(t)

        for file_path in {"foo.py", "bar.pyi", "baz.txt"}:
            (base_dir / file_path).touch()

        command = FakeSingleFileLintCommand(base_dir, FakeSource())
        with mock.patch.object(command, "check", return_value=True) as check:
            reporter = Reporter("fake")
            handler = FakeHandler()
            reporter.process_output.addHandler(handler)
            assert command(reporter) == 0
            assert check.call_count == 2
            assert len(handler.messages) == 0
            assert len(reporter.diagnostics) == 0

        with mock.patch.object(command, "check") as check:
            check.side_effect = [True, False]
            reporter = Reporter("fake")
            handler = FakeHandler()
            reporter.process_output.addHandler(handler)
            assert command(reporter) == 1
            assert check.call_count == 2
            assert len(handler.messages) == 0
            assert len(reporter.diagnostics) == 0


def test_single_file_format_command_base() -> None:
    with TemporaryDirectory() as t:
        base_dir = pathlib.Path(t)

        for file_path in {"foo.py", "bar.pyi", "baz.txt"}:
            (base_dir / file_path).touch()

        command = FakeSingleFileFormatCommand(
            base_dir, FakeSource(), inplace_edit=False
        )
        with mock.patch.object(command, "format", return_value="") as format_method:
            reporter = Reporter("fake")
            handler = FakeHandler()
            reporter.process_output.addHandler(handler)
            assert command(reporter) == 0
            assert format_method.call_count == 2
            assert len(handler.messages) == 0
            assert len(reporter.diagnostics) == 0

        with mock.patch.object(command, "format", return_value="diff") as format_method:
            reporter = Reporter("fake")
            handler = FakeHandler()
            reporter.process_output.addHandler(handler)
            assert command(reporter) == 1
            assert format_method.call_count == 2
            assert len(handler.messages) == 2
            assert len(reporter.diagnostics) == 2
            for file_path in {"foo.py", "bar.pyi"}:
                assert (
                    f"--- {base_dir / file_path}\n"
                    f"+++ {base_dir / file_path}\n"
                    "@@ -0,0 +1 @@\n"
                    "+diff"
                ) in handler.messages
                assert (
                    Diagnostic(
                        start_line=1,
                        end_line=1,
                        start_column=1,
                        file_path=base_dir / file_path,
                        diff="+diff",
                    )
                    in reporter.diagnostics
                )

        command = FakeSingleFileFormatCommand(base_dir, FakeSource(), inplace_edit=True)
        with mock.patch.object(command, "format", return_value=None) as format_method:
            reporter = Reporter("fake")
            handler = FakeHandler()
            reporter.process_output.addHandler(handler)
            assert command(reporter) == 1
            assert format_method.call_count == 2
            assert len(handler.messages) == 0
            assert len(reporter.diagnostics) == 0
            with (base_dir / "foo.py").open() as f:
                assert f.read() == ""
            with (base_dir / "bar.pyi").open() as f:
                assert f.read() == ""

        with mock.patch.object(command, "format", return_value="diff") as format_method:
            reporter = Reporter("fake")
            handler = FakeHandler()
            reporter.process_output.addHandler(handler)
            assert command(reporter) == 0
            assert format_method.call_count == 2
            assert len(handler.messages) == 0
            assert len(reporter.diagnostics) == 0
            with (base_dir / "foo.py").open() as f:
                assert f.read() == "diff"
            with (base_dir / "bar.pyi").open() as f:
                assert f.read() == "diff"



import pathlib
import tempfile
from typing import Sequence

import git
import pytest
from _pytest.monkeypatch import MonkeyPatch

from pysen import git_utils
from pysen.git_utils import GitRepositoryNotFoundError

BASE_DIR = pathlib.Path(__file__).resolve().parent


def test_check_git_available() -> None:
    with tempfile.TemporaryDirectory() as d:
        tempdir = pathlib.Path(d)
        assert git_utils._check_git_enabled()
        assert not git_utils.check_git_available(tempdir)

        git.Repo.init(tempdir)
        assert git_utils._check_git_enabled()
        assert git_utils.check_git_available(tempdir)


def test_skip_git_check(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("PYSEN_IGNORE_GIT", "1")
    with tempfile.TemporaryDirectory() as d:
        tempdir = pathlib.Path(d)
        assert not git_utils._check_git_enabled()
        assert not git_utils.check_git_available(tempdir)

        git.Repo.init(tempdir)
        assert not git_utils._check_git_enabled()
        assert not git_utils.check_git_available(tempdir)


def test_dont_skip_git_check(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("PYSEN_IGNORE_GIT", "0")
    with tempfile.TemporaryDirectory() as d:
        tempdir = pathlib.Path(d)
        assert not git_utils.check_git_available(tempdir)

        git.Repo.init(tempdir)
        assert git_utils.check_git_available(tempdir)


def test_list_indexed_files() -> None:
    def list_indexed_files(target_dir: pathlib.Path) -> Sequence[pathlib.Path]:
        ret = git_utils.list_indexed_files(target_dir)
        git_utils.list_indexed_files.cache_clear()
        return ret

    with tempfile.TemporaryDirectory() as d:
        tempdir = pathlib.Path(d).resolve()

        with pytest.raises(GitRepositoryNotFoundError):
            list(list_indexed_files(tempdir))

        repo = git.Repo.init(tempdir)

        assert list(list_indexed_files(tempdir)) == []

        test_file = tempdir / "hoge"
        test_file.touch()

        assert list(list_indexed_files(tempdir)) == []
        repo.index.add([str(test_file)])

        assert list(list_indexed_files(tempdir)) == [test_file]

        test_dir = tempdir / "foo"
        test_dir2 = tempdir / "foo_2"

        test_dir.mkdir()
        test_dir2.mkdir()

        test_file2 = test_dir / "a"
        test_file2.touch()

        test_file3 = test_dir2 / "b"
        test_file3.touch()

        repo.index.add([str(test_file2), str(test_file3)])
        assert set(list_indexed_files(tempdir)) == {
            test_file,
            test_file2,
            test_file3,
        }
        assert set(list_indexed_files(tempdir / "foo")) == {test_file2}
        assert set(list_indexed_files(tempdir / "foo_2")) == {test_file3}

        assert set(list_indexed_files(tempdir)) == {test_file, test_file2, test_file3}
        # a file is removed without being staged
        test_file3.unlink()
        assert set(list_indexed_files(tempdir)) == {test_file, test_file2}


def test_check_tracked() -> None:
    with tempfile.TemporaryDirectory() as d:
        tempdir = pathlib.Path(d)

        test_file = tempdir / "hoge"
        test_file.touch()

        with pytest.raises(GitRepositoryNotFoundError):
            git_utils.check_tracked(test_file)

        repo = git.Repo.init(tempdir)
        assert not git_utils.check_tracked(test_file)

        repo.index.add([str(test_file)])

        assert git_utils.check_tracked(test_file)



import collections
import pathlib
from typing import DefaultDict, Iterator
from unittest import mock

import pytest

from pysen import mypy
from pysen.mypy import _get_differences_from_base
from pysen.process_utils import add_python_executable
from pysen.reporter import Reporter
from pysen.runner_options import PathContext, RunOptions
from pysen.setting import SettingFile

BASE_DIR = pathlib.Path(__file__).resolve().parent


@pytest.fixture
def reporter() -> Iterator[Reporter]:
    r = Reporter("")
    with r:
        yield r


def test__get_differences_from_base() -> None:
    A = {"A": "a", "B": "b", "C": "c", "X": ["1", "2", "3"], "Y": ["a", "b"]}
    B = {"A": "a", "C": "c2", "D": "d", "X": ["1", "2", "3"], "Y": ["a", "c"]}

    assert _get_differences_from_base(A, B) == {"B": "b", "C": "c", "Y": ["a", "b"]}
    assert _get_differences_from_base(B, A) == {"C": "c2", "D": "d", "Y": ["a", "c"]}


def test_mypy_setting() -> None:
    s = mypy.MypySetting.very_strict()
    assert s == mypy.MypySetting.very_strict()
    assert s != mypy.MypySetting.strict()

    section, settings = s.export(BASE_DIR)
    assert section == ["mypy"]
    assert settings["check_untyped_defs"]
    assert "target_module" not in settings

    s.check_untyped_defs = False
    section, settings = s.export(BASE_DIR)
    assert not settings["check_untyped_defs"]

    section, settings = s.export(BASE_DIR, target_module="hoge.fuga.*")
    assert section == ["mypy-hoge.fuga.*"]

    s.mypy_path = [
        "/opt/pysen/stubs",
        "stubs2",
        pathlib.Path("/usr/pysen/stubs3"),
        pathlib.Path("stub4"),
    ]
    section, settings = s.export(pathlib.Path("/opt/pysen/package/python"))
    assert settings["mypy_path"] == [
        "/opt/pysen/stubs",
        "stubs2",
        "/usr/pysen/stubs3",
        "stub4",
    ]

    # This option is set by pyproject loader
    s._pysen_convert_abspath = True
    section, settings = s.export(pathlib.Path("/opt/pysen/package/python"))
    assert settings["mypy_path"] == [
        "../../stubs",
        "stubs2",
        "../../../../usr/pysen/stubs3",
        "stub4",
    ]


def test_settings() -> None:
    m = mypy.Mypy(
        setting=mypy.MypySetting.very_strict(),
        module_settings={"hoge.fuga": mypy.MypySetting.strict()},
    )
    assert m.setting == mypy.MypySetting.very_strict()
    assert m.module_settings == {"hoge.fuga": mypy.MypySetting.strict()}

    m = mypy.Mypy()
    assert m.setting == mypy.MypySetting()
    assert m.module_settings == {}


def test_commands(reporter: Reporter) -> None:
    m = mypy.Mypy(
        mypy_targets=[mypy.MypyTarget([pathlib.Path("/bar"), pathlib.Path("baz")])]
    )
    expected_cmds = add_python_executable(
        "mypy",
        "--show-absolute-path",
        "--no-color-output",
        "--show-column-numbers",
        "--no-error-summary",
        "--config-file",
        "/setting/setup.cfg",
        "/bar",
        "/foo/baz",
    )
    cmd = m.create_command(
        "lint",
        PathContext(pathlib.Path("/foo"), pathlib.Path("/setting")),
        RunOptions(),
    )

    with mock.patch("os.chdir", return_value=None):
        with mock.patch("pysen.process_utils.run", return_value=(0, "", "")) as patch:
            assert cmd(reporter=reporter) == 0
            patch.assert_called_with(expected_cmds, reporter)


def test_export_settings() -> None:
    m = mypy.Mypy(
        setting=mypy.MypySetting(
            mypy_path=["hoge"],
            plugins=[mypy.MypyPlugin(script=BASE_DIR / pathlib.Path("foo/bar"))],
            disallow_any_decorated=False,
            ignore_missing_imports=False,
            warn_redundant_casts=True,
            follow_imports=mypy.MypyFollowImports.ERROR,
            _pysen_convert_abspath=True,
        ),
        module_settings={
            "foo.*": mypy.MypySetting(disallow_any_decorated=True),  # duplicated
            "bar.baz": mypy.MypySetting(
                ignore_missing_imports=True,  # duplicated
                disallow_any_decorated=False,  # same (not emitted in exported settings)
                disallow_any_unimported=False,  # new
            ),
        },
    )
    files: DefaultDict[str, SettingFile] = collections.defaultdict(SettingFile)
    m.export_settings(PathContext(BASE_DIR, BASE_DIR), files)

    assert files.keys() == {"setup.cfg"}
    setting_file = files["setup.cfg"]
    expected = {
        "mypy": {
            "disallow_any_decorated": False,
            "follow_imports": "error",
            "ignore_missing_imports": False,
            "mypy_path": ["hoge"],
            "warn_redundant_casts": True,
            "plugins": ["foo/bar"],
        },
        "mypy-foo.*": {"disallow_any_decorated": True},
        "mypy-bar.baz": {
            "disallow_any_unimported": False,
            "ignore_missing_imports": True,
        },
    }

    assert setting_file.as_dict() == expected



from unittest import mock

from pysen.cli import _use_pretty_logging


def test__use_pretty_logging() -> None:
    with mock.patch("sys.stderr.isatty", return_value=True):
        assert _use_pretty_logging()

    with mock.patch("sys.stderr.isatty", return_value=False):
        assert not _use_pretty_logging()



import pathlib
import subprocess
import tempfile
import unittest.mock
from typing import Iterator, List

import pytest

from pysen.command import check_command_installed
from pysen.exceptions import CommandNotFoundError


@pytest.fixture
def invalid_command() -> Iterator[List[str]]:
    with tempfile.TemporaryDirectory() as td:
        yield [str(pathlib.Path(td) / "this_command_does_not_exist")]


def test_check_command_installed(invalid_command: List[str]) -> None:
    check_command_installed("echo", "a")

    with pytest.raises(CommandNotFoundError):
        check_command_installed(*invalid_command)
    with unittest.mock.patch("subprocess.call", return_value=127):
        with pytest.raises(CommandNotFoundError):
            assert subprocess.call("hoge") == 127
            check_command_installed("hoge")



import enum
import pathlib
from typing import DefaultDict, Dict, List, Optional, Sequence, Tuple

from pysen import ComponentBase
from pysen.command import CommandBase
from pysen.diagnostic import Diagnostic
from pysen.reporter import Reporter
from pysen.runner_options import PathContext, RunOptions
from pysen.setting import SettingFile


class Operation(enum.Enum):
    ADD = "+"
    MUL = "*"


class FakeCommand(CommandBase):
    def __init__(
        self, coef: int, op: Operation, ref: List[float], options: RunOptions
    ) -> None:
        self.coef = coef
        self.op = op
        self.ref = ref
        self.options = options
        assert len(ref) == 1

    @property
    def name(self) -> str:
        return f"{self.op.value} {self.coef}"

    def __call__(self, reporter: Reporter) -> int:
        value = self.ref[0]
        coef = float(self.coef)

        if self.op == Operation.ADD:
            value += coef
        elif self.op == Operation.MUL:
            value *= coef
        else:
            raise AssertionError(f"invalid op: {self.op}")

        self.ref[0] = value

        if value >= 0.0:
            return 0
        else:
            if self.options.require_diagnostics:
                reporter.report_diagnostics(
                    [Diagnostic(pathlib.Path(".").resolve(), message="")]
                )
            return 1


class FakeComponent(ComponentBase):
    def __init__(
        self,
        name: str,
        ops: Dict[str, Tuple[int, Operation]],
        expected_base_dir: Optional[pathlib.Path],
        expected_settings_dir: Optional[pathlib.Path],
        ref: List[float],
    ) -> None:
        self._name = name
        self._ops = ops
        self._expected_base_dir = expected_base_dir
        self._expected_settings_dir = expected_settings_dir
        self._ref = ref
        assert len(ref) == 1

    @property
    def name(self) -> str:
        return self._name

    def export_settings(
        self,
        paths: PathContext,
        files: DefaultDict[str, SettingFile],
    ) -> None:
        if self._expected_base_dir is not None:
            assert paths.base_dir == self._expected_base_dir
        if self._expected_settings_dir is not None:
            assert paths.settings_dir == self._expected_settings_dir
        for name, op in self._ops.items():
            fname = f"{name}.yaml"
            setting_file = files[fname]
            setting_file.set_section((self.name,), {"coef": op[0], "op": op[1].value})

    @property
    def targets(self) -> Sequence[str]:
        return list(self._ops.keys())

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        if self._expected_base_dir is not None:
            assert paths.base_dir == self._expected_base_dir
        if self._expected_settings_dir is not None:
            assert paths.settings_dir == self._expected_settings_dir
        op = self._ops[target]
        return FakeCommand(op[0], op[1], self._ref, options)



import argparse
import pathlib
from typing import Any, Callable, Dict, List, Optional, Sequence

import pytest

from fakes.component import FakeComponent, Operation
from pysen import dumper
from pysen.component import ComponentBase
from pysen.exceptions import InvalidComponentName
from pysen.manifest import Manifest, export_settings, get_target, get_targets
from pysen.runner_options import PathContext, RunOptions
from pysen.setting import SettingFile

FixtureType = Callable[..., Sequence[ComponentBase]]


@pytest.fixture
def fake_components() -> FixtureType:
    def create(
        base_dir: Optional[pathlib.Path],
        settings_dir: Optional[pathlib.Path],
        ref: Optional[List[float]] = None,
    ) -> Sequence[ComponentBase]:
        r = ref or [1.0]
        return (
            FakeComponent(
                "node1",
                {"op1": (2, Operation.MUL), "op2": (10, Operation.ADD)},
                base_dir,
                settings_dir,
                r,
            ),
            FakeComponent(
                "node2",
                {"op1": (3, Operation.MUL), "op3": (-1, Operation.MUL)},
                base_dir,
                settings_dir,
                r,
            ),
        )

    return create


def test_components(fake_components: FixtureType) -> None:
    components = fake_components(pathlib.Path(), pathlib.Path())
    c = Manifest(components)
    assert c.components == list(components)


def test_default_dump_handler() -> None:
    m = Manifest()
    assert m._dump_handler == dumper.dump


def test_export_settings(fake_components: FixtureType) -> None:
    base_dir = pathlib.Path("/foo/bar")
    settings_dir = pathlib.Path("/settings")
    paths = PathContext(base_dir, settings_dir)
    components = fake_components(base_dir, settings_dir)

    dumped: Dict[str, Dict[str, Any]] = {}

    def dump(s_dir: pathlib.Path, fname: str, data: SettingFile) -> None:
        assert s_dir == settings_dir
        dumped[fname] = data.as_dict()

    export_settings(paths, components, dump)
    assert len(dumped.keys()) == 3
    expected = {
        "op1.yaml": {"node1": {"coef": 2, "op": "*"}, "node2": {"coef": 3, "op": "*"}},
        "op2.yaml": {"node1": {"coef": 10, "op": "+"}},
        "op3.yaml": {"node2": {"coef": -1, "op": "*"}},
    }
    assert expected == dumped
    dumped.clear()

    m = Manifest(components, dump_handler=dump)
    m.export_settings(paths, argparse.Namespace(enable=None, disable=None))
    assert len(dumped.keys()) == 3
    expected = {
        "op1.yaml": {"node1": {"coef": 2, "op": "*"}, "node2": {"coef": 3, "op": "*"}},
        "op2.yaml": {"node1": {"coef": 10, "op": "+"}},
        "op3.yaml": {"node2": {"coef": -1, "op": "*"}},
    }
    assert expected == dumped

    dumped.clear()
    # NOTE(igarashi): export_settings intentionally ignores enable and disable arguments.
    # See manifest.py for more details
    m.export_settings(paths, argparse.Namespace(enable=["node1"], disable=None))
    assert len(dumped.keys()) == 3
    expected = {
        "op1.yaml": {"node1": {"coef": 2, "op": "*"}, "node2": {"coef": 3, "op": "*"}},
        "op2.yaml": {"node1": {"coef": 10, "op": "+"}},
        "op3.yaml": {"node2": {"coef": -1, "op": "*"}},
    }
    assert expected == dumped


def test_get_targets(fake_components: FixtureType) -> None:
    components = fake_components(None, None)

    targets = get_targets(components)
    assert targets == {"op1": ["node1", "node2"], "op2": ["node1"], "op3": ["node2"]}

    m = Manifest(components)
    targets = m.get_targets(argparse.Namespace(enable=None, disable=None))
    assert targets == {"op1": ["node1", "node2"], "op2": ["node1"], "op3": ["node2"]}

    targets = m.get_targets(argparse.Namespace(enable=None, disable=["node1"]))
    assert targets == {"op1": ["node2"], "op3": ["node2"]}

    targets = m.get_targets(argparse.Namespace(enable=["node1"], disable=None))
    assert targets == {"op1": ["node1"], "op2": ["node1"]}

    with pytest.raises(InvalidComponentName) as e:
        m.get_targets(argparse.Namespace(enable=["noexist", "ditto"], disable=None))
    assert (
        "The following component(s) in option --enable were not found: ditto,noexist"
        in str(e)
    )
    with pytest.raises(InvalidComponentName) as e:
        m.get_targets(argparse.Namespace(disable=["ditto"], enable=None))
    assert (
        "The following component(s) in option --disable were not found: ditto" in str(e)
    )


def test_get_target(fake_components: FixtureType) -> None:
    base_dir = pathlib.Path("/foo/bar")
    settings_dir = pathlib.Path("/settings")
    paths = PathContext(base_dir, settings_dir)
    options = RunOptions()
    components = fake_components(base_dir, settings_dir)

    target = get_target("op1", components, paths, options)
    assert len(target) == 2
    assert {target[0].name, target[1].name} == {"* 2", "* 3"}

    assert get_target("op99", components, paths, options) == []

    m = Manifest(components)

    # expected to be same as get_target("op1", component, base_dir)
    target = m.get_target(
        "op1", paths, options, argparse.Namespace(enable=None, disable=None)
    )
    assert len(target) == 2
    assert {target[0].name, target[1].name} == {"* 2", "* 3"}

    target = m.get_target(
        "op2", paths, options, argparse.Namespace(enable=None, disable=None)
    )
    assert len(target) == 1
    assert {target[0].name} == {"+ 10"}

    assert (
        m.get_target(
            "op99", paths, options, argparse.Namespace(enable=None, disable=None)
        )
        == []
    )

    # check if components are filtered by disable option
    target = m.get_target(
        "op1", paths, options, argparse.Namespace(enable=None, disable=["node1"])
    )
    assert len(target) == 1
    assert {target[0].name} == {"* 3"}

    assert (
        m.get_target(
            "op2", paths, options, argparse.Namespace(enable=None, disable=["node1"])
        )
        == []
    )

    target = m.get_target(
        "op3", paths, options, argparse.Namespace(enable=None, disable=["node1"])
    )
    assert len(target) == 1
    assert target[0].name == "* -1"



import math
import pathlib
import threading
import unittest.mock
from typing import Any, Callable, List, Optional, cast

import pytest
from _pytest.capture import CaptureFixture

from fakes.component import FakeCommand, FakeComponent, Operation
from fakes.manifest import FakeManifest
from pysen.command import CommandBase
from pysen.exceptions import CommandNotFoundError, InvalidCommandNameError
from pysen.manifest import Manifest, ManifestBase
from pysen.reporter import Reporter, ReporterFactory
from pysen.runner import Runner, _has_side_effects, _verify_command_name
from pysen.runner_options import PathContext, RunOptions

FAKE_PATH = pathlib.Path(__file__)
FixtureType = Callable[..., ManifestBase]


@pytest.fixture
def fake_manifest() -> FixtureType:
    def create(
        base_dir: pathlib.Path,
        settings_dir: Optional[pathlib.Path],
        ref: Optional[List[float]] = None,
        dumped: Optional[threading.Event] = None,
    ) -> ManifestBase:
        r = ref or [0.0]
        d = dumped or threading.Event()

        node1 = FakeComponent(
            "node1",
            {"op1": (2, Operation.MUL), "op2": (10, Operation.ADD)},
            base_dir,
            settings_dir,
            r,
        )
        node2 = FakeComponent(
            "node2",
            {"op1": (3, Operation.MUL), "op3": (-1, Operation.MUL)},
            base_dir,
            settings_dir,
            r,
        )

        def dump(*args: Any) -> None:
            d.set()

        return Manifest([node1, node2], dump_handler=dump)

    return create


@pytest.fixture
def fake_manifest_with_options() -> FixtureType:
    def create(
        base_dir: pathlib.Path,
        settings_dir: Optional[pathlib.Path],
        num_required: bool,
        ref: Optional[List[float]] = None,
    ) -> ManifestBase:
        r = ref or [0.0]
        options = RunOptions()

        items = {
            "op1": [
                FakeCommand(2, Operation.ADD, r, options),
                FakeCommand(10, Operation.MUL, r, options),
                FakeCommand(-1, Operation.ADD, r, options),
            ]
        }
        special_item = [
            FakeCommand(3, Operation.ADD, r, options),
            FakeCommand(2, Operation.MUL, r, options),
        ]

        ret = FakeManifest(base_dir, settings_dir, num_required, items, special_item)
        return cast(ManifestBase, ret)

    return create


def test_export_settings(fake_manifest: FixtureType) -> None:
    base_dir = pathlib.Path("/foo")
    settings_dir = pathlib.Path("/bar")
    dumped = threading.Event()
    manifest = fake_manifest(base_dir, settings_dir, dumped=dumped)
    runner = Runner(manifest)
    assert not dumped.is_set()
    args = runner.parse_manifest_arguments([])
    runner.export_settings(base_dir, settings_dir, args)
    assert dumped.is_set()


def test_get_targets(fake_manifest: FixtureType) -> None:
    paths = PathContext(pathlib.Path("/foo"), pathlib.Path("/bar"))
    manifest = fake_manifest(paths.base_dir, paths.settings_dir)
    runner = Runner(manifest)
    args = runner.parse_manifest_arguments([])
    targets = runner.get_targets(args)
    assert len(targets) == 3
    assert targets.keys() == {"op1", "op2", "op3"}
    assert len(targets["op1"]) == 2
    assert len(targets["op2"]) == 1
    assert len(targets["op3"]) == 1


def test_get_target(fake_manifest: FixtureType) -> None:
    paths = PathContext(pathlib.Path("/foo"), pathlib.Path("/bar"))
    options = RunOptions()
    manifest = fake_manifest(paths.base_dir, paths.settings_dir)
    runner = Runner(manifest)
    args = runner.parse_manifest_arguments([])
    assert len(runner._get_target("op1", paths, options, args)) == 2
    assert len(runner._get_target("op2", paths, options, args)) == 1
    assert len(runner._get_target("op3", paths, options, args)) == 1
    with pytest.raises(CommandNotFoundError):
        runner._get_target("op4", paths, options, args)


def test_parse_manifest_arguments(
    fake_manifest_with_options: FixtureType, capsys: CaptureFixture[str]
) -> None:
    ref = [1.0]
    manifest = fake_manifest_with_options(
        pathlib.Path("/foo"), None, num_required=False, ref=ref
    )
    runner = Runner(manifest)

    parsed = runner.parse_manifest_arguments([])
    assert parsed.num == 0
    assert not parsed.special

    parsed = runner.parse_manifest_arguments(["--special"])
    assert parsed.num == 0
    assert parsed.special

    parsed = runner.parse_manifest_arguments(["--special", "--num", "5"])
    assert parsed.num == 5
    assert parsed.special

    capsys.readouterr()
    with pytest.raises(SystemExit):
        runner.parse_manifest_arguments(["--hoge"])
    out = capsys.readouterr()
    assert "error: unrecognized arguments" in out.err

    manifest = fake_manifest_with_options(
        pathlib.Path("/foo"), None, num_required=True, ref=ref
    )
    runner = Runner(manifest)

    capsys.readouterr()
    with pytest.raises(SystemExit):
        runner.parse_manifest_arguments([])
    out = capsys.readouterr()
    assert "the following arguments are required" in out.err

    parsed = runner.parse_manifest_arguments(["--num", "2"])
    assert parsed.num == 2
    assert not parsed.special


def test_run_with_settings_dir(fake_manifest: FixtureType) -> None:
    base_dir = pathlib.Path("/foo")
    settings_dir = pathlib.Path("/settings")
    options = RunOptions()
    ref = [1.0]
    manifest = fake_manifest(base_dir, settings_dir, ref)
    runner = Runner(manifest)
    reporters = ReporterFactory()

    manifest_args = runner.parse_manifest_arguments([])
    with unittest.mock.patch("pathlib.Path.mkdir") as mock:
        runner.run(
            "op1",
            base_dir,
            manifest_args,
            reporters,
            options,
            settings_dir=settings_dir,
            files=None,
        )
        mock.assert_called()
    assert math.isclose(ref[0], 6.0)
    assert not reporters.has_error()


def test_run(fake_manifest: FixtureType) -> None:
    base_dir = pathlib.Path("/foo")
    ref = [1.0]
    manifest = fake_manifest(base_dir, None, ref)
    runner = Runner(manifest)
    reporters = ReporterFactory()
    options = RunOptions()

    manifest_args = runner.parse_manifest_arguments([])
    runner.run(
        "op1",
        base_dir,
        manifest_args=manifest_args,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 6.0)
    assert not reporters.has_error()

    ref[0] = 1.0
    runner.run(
        "op2",
        base_dir,
        manifest_args=manifest_args,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 11.0)
    assert not reporters.has_error()

    ref[0] = 1.0
    runner.run(
        "op3",
        base_dir,
        manifest_args=manifest_args,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], -1.0)
    assert reporters.has_error()
    reporter = next(r for r in reporters._reporters if r.name == "* -1")
    assert len(reporter.diagnostics) == 1

    # check if options is correctly handled in Command through runner.run
    reporters = ReporterFactory()
    options = RunOptions(require_diagnostics=False)
    ref[0] = 1.0
    runner.run(
        "op3",
        base_dir,
        manifest_args=manifest_args,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], -1.0)
    assert reporters.has_error()
    reporter = next(r for r in reporters._reporters if r.name == "* -1")
    assert len(reporter.diagnostics) == 0

    reporters = ReporterFactory()

    with pytest.raises(CommandNotFoundError):
        runner.run(
            "op4",
            base_dir,
            manifest_args=manifest_args,
            reporters=reporters,
            options=options,
            files=None,
        )

    ref[0] = 1.0
    manifest_args = runner.parse_manifest_arguments(["--enable", "node1"])
    runner.run(
        "op1",
        base_dir,
        manifest_args=manifest_args,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 2.0)


def test_run_manifest_args(fake_manifest_with_options: FixtureType) -> None:
    base_dir = pathlib.Path("/foo")
    ref = [1.0]
    manifest = fake_manifest_with_options(base_dir, None, num_required=False, ref=ref)
    runner = Runner(manifest)
    reporters = ReporterFactory()
    options = RunOptions()

    parsed = runner.parse_manifest_arguments([])
    runner.run(
        "op1",
        base_dir,
        manifest_args=parsed,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 29.0)
    assert not reporters.has_error()

    ref[0] = 1.0
    parsed = runner.parse_manifest_arguments([])
    with pytest.raises(CommandNotFoundError):
        runner.run(
            "special",
            base_dir,
            manifest_args=parsed,
            reporters=reporters,
            options=options,
            files=None,
        )

    ref[0] = 1.0
    parsed = runner.parse_manifest_arguments(["--special"])
    runner.run(
        "special",
        base_dir,
        manifest_args=parsed,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 8.0)
    assert not reporters.has_error()

    ref[0] = 1.0
    parsed = runner.parse_manifest_arguments(["--num", "1", "--special"])
    runner.run(
        "special",
        base_dir,
        manifest_args=parsed,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 4.0)
    assert not reporters.has_error()

    manifest = fake_manifest_with_options(base_dir, None, num_required=True, ref=ref)
    runner = Runner(manifest)

    ref[0] = 1.0
    parsed = runner.parse_manifest_arguments(["--num", "1"])
    runner.run(
        "op1",
        base_dir,
        manifest_args=parsed,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 3.0)
    assert not reporters.has_error()

    ref[0] = 1.0
    parsed = runner.parse_manifest_arguments(["--num", "2"])
    runner.run(
        "op1",
        base_dir,
        manifest_args=parsed,
        reporters=reporters,
        options=options,
        files=None,
    )
    assert math.isclose(ref[0], 30.0)
    assert not reporters.has_error()


class ValidCommand(CommandBase):
    @property
    def name(self) -> str:
        return "valid"

    def __call__(self, reporter: Reporter) -> int:
        return 0


class InvalidCommand(CommandBase):
    @property
    def name(self) -> str:
        return "my:py"

    def __call__(self, reporter: Reporter) -> int:
        return 0


def test__verify_command_name() -> None:
    _verify_command_name(ValidCommand())
    with pytest.raises(InvalidCommandNameError):
        _verify_command_name(InvalidCommand())


class MockCommand(CommandBase):
    @property
    def name(self) -> str:
        "mock"

    def __call__(self, reporter: Reporter) -> int:
        return 0

    @property
    def has_side_effects(self) -> bool:
        return True


class PurelyFunctionalCommand(CommandBase):
    @property
    def name(self) -> str:
        "preferred_command"

    def __call__(self, reporter: Reporter) -> int:
        return 0

    @property
    def has_side_effects(self) -> bool:
        return False


def test__has_side_effects() -> None:
    pfc = PurelyFunctionalCommand()
    mc = MockCommand()
    assert _has_side_effects([pfc, pfc, pfc, mc])
    assert _has_side_effects([mc, mc, mc, mc])
    assert not _has_side_effects([pfc, pfc, pfc, pfc])



import dataclasses
import pathlib
import subprocess
from typing import DefaultDict, List, Sequence

import dacite

from pysen.command import CommandBase
from pysen.component import ComponentBase, RunOptions
from pysen.path import change_dir
from pysen.plugin import PluginBase
from pysen.pyproject_model import Config, PluginConfig
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.setting import SettingFile


class ShellCommand(CommandBase):
    def __init__(self, name: str, base_dir: pathlib.Path, cmd: Sequence[str]) -> None:
        self._name = name
        self._base_dir = base_dir
        self._cmd = cmd

    @property
    def name(self) -> str:
        return self._name

    def __call__(self, reporter: Reporter) -> int:
        with change_dir(self._base_dir):
            try:
                ret = subprocess.run(self._cmd)
                reporter.logger.info(f"{self._cmd} returns {ret.returncode}")
                return ret.returncode
            except BaseException as e:
                reporter.logger.info(
                    f"an error occurred while executing: {self._cmd}\n{e}"
                )
                return 255


class ShellComponent(ComponentBase):
    def __init__(self, name: str, cmd: Sequence[str], targets: Sequence[str]) -> None:
        self._name = name
        self._cmd = cmd
        self._targets = targets

    @property
    def name(self) -> str:
        return self._name

    def export_settings(
        self, paths: PathContext, files: DefaultDict[str, SettingFile],
    ) -> None:
        print(f"Called export_settings at {self._name}: do nothing")

    @property
    def targets(self) -> Sequence[str]:
        return self._targets

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        assert target in self._targets
        return ShellCommand(self._name, paths.base_dir, self._cmd)


@dataclasses.dataclass
class ShellPluginConfig:
    name: str
    command: List[str]
    targets: List[str]


class ShellPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config_data: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        assert (
            config_data.config is not None
        ), f"{config_data.location}.config must be not None"
        config = dacite.from_dict(
            ShellPluginConfig, config_data.config, dacite.Config(strict=True)
        )
        return [ShellComponent(config.name, config.command, config.targets)]


# NOTE(igarashi): This is the entry point of a plugin method
def plugin() -> PluginBase:
    return ShellPlugin()



import pytest

from pysen.py_version import PythonVersion, VersionRepresentation


def test_python_version() -> None:
    py38 = PythonVersion(3, 8)

    assert py38 == PythonVersion(3, 8)
    assert py38 != PythonVersion(3, 9)

    assert py38.version == "3.8"
    assert py38.full_representation == "Python3.8"
    assert py38.short_representation == "py38"

    py388 = PythonVersion(3, 8, 8)

    assert py388 == PythonVersion(3, 8, 8)
    assert py388 != PythonVersion(3, 8, 9)
    assert py388 != py38

    assert py388.version == "3.8.8"
    assert py388.full_representation == "Python3.8.8"
    assert py388.short_representation == "py38"


def test_version_ops() -> None:
    assert VersionRepresentation(3, 6) == VersionRepresentation(3, 6, None, None)
    with pytest.raises(NotImplementedError):
        assert VersionRepresentation(3, 6) == "3"


def test_version_comp() -> None:
    def assert_rhs_is_larger(
        lhs: VersionRepresentation, rhs: VersionRepresentation
    ) -> None:
        assert lhs < rhs
        assert not rhs < lhs

    assert_rhs_is_larger(VersionRepresentation(0, 4, 5), VersionRepresentation(1, 2, 3))
    assert_rhs_is_larger(VersionRepresentation(1, 0), VersionRepresentation(2, 0))
    assert_rhs_is_larger(VersionRepresentation(0, 5), VersionRepresentation(0, 6))
    assert_rhs_is_larger(VersionRepresentation(0, 5), VersionRepresentation(0, 5, 1))

    assert not VersionRepresentation(0, 5, 1, "b0") < VersionRepresentation(
        0, 5, 1, "b1"
    )
    assert not VersionRepresentation(0, 5, 1, "b1") < VersionRepresentation(
        0, 5, 1, "b0"
    )


def test_is_compatible() -> None:
    assert VersionRepresentation(0, 5).is_compatible(VersionRepresentation(0, 5))
    assert VersionRepresentation(0, 5).is_compatible(VersionRepresentation(0, 5, 7))
    assert VersionRepresentation(0, 5).is_compatible(VersionRepresentation(0, 5, 1))
    assert VersionRepresentation(0, 5, 5).is_compatible(VersionRepresentation(0, 5, 4))
    assert VersionRepresentation(0, 5, 5).is_compatible(VersionRepresentation(0, 5, 6))
    assert VersionRepresentation(3, 6).is_compatible(VersionRepresentation(3, 6))
    assert not VersionRepresentation(3, 6).is_compatible(VersionRepresentation(3, 5))
    assert not VersionRepresentation(3, 6, 8).is_compatible(
        VersionRepresentation(3, 5, 32)
    )
    assert not VersionRepresentation(2, 6).is_compatible(VersionRepresentation(3, 0))
    assert not VersionRepresentation(4, 6).is_compatible(VersionRepresentation(3, 0))


def test_version_from_str() -> None:
    def check_version(s: str, expected: VersionRepresentation) -> None:
        actual = VersionRepresentation.from_str(s)
        assert actual == expected
        assert s == str(expected)

    cases = {
        ("0.601", VersionRepresentation(0, 601)),
        ("3.0.8", VersionRepresentation(3, 0, 8)),
        ("3.6.8a1", VersionRepresentation(3, 6, 8, "a1")),
        ("3.6a1", VersionRepresentation(3, 6, None, "a1")),
        ("3.6b0", VersionRepresentation(3, 6, None, "b0")),
        ("3.6rc993", VersionRepresentation(3, 6, None, "rc993")),
    }
    for case in cases:
        check_version(*case)

    with pytest.raises(ValueError):
        # prelease phase must be either a, b, or rc
        VersionRepresentation.from_str("3.6.8alpha1")
    with pytest.raises(ValueError):
        # MUST have a pre-release number
        VersionRepresentation.from_str("3.6.8a")
    with pytest.raises(ValueError):
        # MUST NOT start with zero followed by another number
        VersionRepresentation.from_str("03.1")
    with pytest.raises(ValueError):
        # MUST NOT start with zero followed by another number
        VersionRepresentation.from_str("00.1")
    with pytest.raises(ValueError):
        # MUST have minor
        VersionRepresentation.from_str("3")
    with pytest.raises(ValueError):
        # too many dots
        VersionRepresentation.from_str("3.0.100.1")
    with pytest.raises(ValueError):
        VersionRepresentation.from_str("3.")



import pathlib
import tempfile
from typing import List

import pytest

from pysen import ManifestBase
from pysen.black import Black
from pysen.component import ComponentBase
from pysen.exceptions import InvalidManifestBuilderError
from pysen.flake8 import Flake8
from pysen.manifest import Manifest
from pysen.manifest_builder import _build, _build_external, build
from pysen.source import Source

CURRENT_FILE = pathlib.Path(__file__).resolve()
BASE_DIR = CURRENT_FILE.parent


def get_source_from_good_builder(manifest: ManifestBase) -> Source:
    assert isinstance(manifest, Manifest)
    flake8 = manifest.get_component("flake8")
    assert isinstance(flake8, Flake8)
    return flake8.source


def test__build_external() -> None:
    manifest = _build_external(
        BASE_DIR / "fakes/configs/good_builder.py", [], CURRENT_FILE
    )
    assert get_source_from_good_builder(manifest).includes.keys() == {BASE_DIR}

    test_source = Source(includes=["/hoge/fuga"])
    manifest = _build_external(
        BASE_DIR / "fakes/configs/good_builder.py",
        [Black(source=test_source)],
        CURRENT_FILE,
    )
    assert get_source_from_good_builder(manifest) == test_source

    with tempfile.TemporaryDirectory() as d:
        tempdir = pathlib.Path(d)
        manifest = _build_external(
            BASE_DIR / "fakes/configs/good_builder.py", [], tempdir / "hoge"
        )
        assert get_source_from_good_builder(manifest).includes.keys() == {tempdir}

    with pytest.raises(InvalidManifestBuilderError) as ex:
        _build_external(
            BASE_DIR / "fakes/configs/invalid_interface_builder.py", [], CURRENT_FILE
        )

    assert "external builder must have" in str(ex.value)

    with pytest.raises(InvalidManifestBuilderError) as ex:
        _build_external(
            BASE_DIR / "fakes/configs/invalid_return_builder.py", [], CURRENT_FILE
        )

    assert "instance of ManifestBase" in str(ex.value)


def test__build() -> None:
    manifest = _build([], CURRENT_FILE)
    assert isinstance(manifest, Manifest)
    assert len(manifest.components) == 0

    black = Black()
    manifest = _build([black], CURRENT_FILE)
    assert isinstance(manifest, Manifest)
    assert manifest.components == [black]


def test_build() -> None:
    test_source = Source(includes=["/hoge/fuga"])
    test_components: List[ComponentBase] = [Black(source=test_source)]

    manifest = build(test_components, CURRENT_FILE)
    assert isinstance(manifest, Manifest)
    assert manifest.components == test_components

    manifest = build(
        [],
        CURRENT_FILE,
        external_builder=BASE_DIR / "fakes/configs/good_builder.py",
    )
    assert isinstance(manifest, Manifest)
    assert len(manifest.components) == 2
    assert get_source_from_good_builder(manifest).includes.keys() == {BASE_DIR}

    manifest = build(
        test_components,
        CURRENT_FILE,
        external_builder=BASE_DIR / "fakes/configs/good_builder.py",
    )
    assert isinstance(manifest, Manifest)
    assert len(manifest.components) == 2
    assert get_source_from_good_builder(manifest) == test_source



import pathlib
from typing import Optional, Sequence

from pysen import ComponentBase, Manifest, ManifestBase, Source, factory
from pysen.black import Black


def build(
    components: Sequence[ComponentBase], src_path: Optional[pathlib.Path]
) -> ManifestBase:
    black = next((x for x in components if isinstance(x, Black)), None)
    source: Source

    if black is not None:
        source = black.source
    else:
        src_path = src_path or pathlib.Path.cwd()
        source = Source(includes=[src_path.resolve().parent])

    components = factory.configure_lint(
        factory.ConfigureLintOptions(
            enable_flake8=True,
            enable_isort=True,
            source=source,
        )
    )
    return Manifest(components)



import pathlib

import pytest

BASE_DIR = pathlib.Path(__file__).resolve().parent


@pytest.fixture
def example_dir() -> pathlib.Path:
    ret = BASE_DIR.parents[1] / "examples"
    assert ret.exists()
    return ret



import pathlib
from typing import Optional, Sequence

import pysen
from pysen import Source
from pysen.component import ComponentBase
from pysen.manifest import Manifest, ManifestBase


def build(
    components: Sequence[ComponentBase], src_path: Optional[pathlib.Path]
) -> ManifestBase:
    source = Source([".", "tests", "tools"])
    mypy = pysen.Mypy(
        setting=pysen.MypySetting.strict(),
        module_settings={  # per module setting
            "example_advanced_package.cmds.*": pysen.MypySetting.entry(),
            "example_advanced_package.tools.*": pysen.MypySetting(
                disallow_any_generics=False,
            ),
        },
        mypy_targets=[
            pysen.MypyTarget([pathlib.Path(".")]),
            pysen.MypyTarget([pathlib.Path("tools")]),
        ],
    )
    flake8_setting = pysen.Flake8Setting(
        ignore=["W", "E"], select=["B", "C", "F"], max_complexity=10
    )
    flake8 = pysen.Flake8(setting=flake8_setting, source=source)

    return Manifest([mypy, flake8])



import pytest
import tomlkit

from pysen.pyproject_model import _get_descendant
from pysen.setting import _create_dict, _traverse_toml


def test__create_dict() -> None:
    assert _create_dict([]) == {}
    assert _create_dict(["foo"]) == {"foo": {}}
    assert _create_dict(["foo", "bar"]) == {"foo": {"bar": {}}}
    assert _create_dict(["foo", "bar", "baz"]) == {"foo": {"bar": {"baz": {}}}}


def test__traverse_toml() -> None:
    unordered_document = """
    [tool.poetry]
    hoo=1
    [build-system]
    foo=42
    [tool.pysen-cli]
    bar=43
    """
    document = tomlkit.loads(unordered_document)
    tool = document["tool"]
    # OutOfOrderTableProxy is problematic as it is a reference and cannot be traversed.
    assert isinstance(tool, tomlkit.container.OutOfOrderTableProxy)
    with pytest.raises(ValueError):
        _traverse_toml(("tool", "hoge", "answer"), document, False)
    _traverse_toml(("tool", "hoge", "answer"), document, True)
    assert _get_descendant(document, ["tool", "hoge", "answer"]) == {}



import argparse
import pathlib
from typing import Dict, List, Optional, Sequence

from pysen import dumper
from pysen.component import ComponentBase, LintComponentBase
from pysen.manifest import (
    ComponentName,
    ManifestBase,
    ParserType,
    TargetName,
    TargetType,
    export_settings,
    get_target,
    get_targets,
)
from pysen.runner_options import PathContext, RunOptions


class CustomManifest(ManifestBase):
    def __init__(self, components: Sequence[ComponentBase]) -> None:
        self._components = components

    def configure_parser(self, parser: ParserType) -> None:
        super().configure_parser(parser)

        parser.add_argument(
            "--ignore-lint",
            action="store_true",
            help="set True to ignore lint components",
        )

    def _get_components(self, args: argparse.Namespace) -> Sequence[ComponentBase]:
        components = self._components
        if args.ignore_lint:
            components = [c for c in components if not isinstance(c, LintComponentBase)]
        return components

    def export_settings(self, paths: PathContext, args: argparse.Namespace) -> None:
        export_settings(paths, self._components, dumper.dump)

    def get_targets(self, args: argparse.Namespace) -> Dict[str, List[ComponentName]]:
        return get_targets(self._get_components(args))

    def get_target(
        self,
        target: TargetName,
        paths: PathContext,
        options: RunOptions,
        args: argparse.Namespace,
    ) -> TargetType:
        return get_target(target, self._get_components(args), paths, options)


def build(
    components: Sequence[ComponentBase], src_path: Optional[pathlib.Path]
) -> ManifestBase:
    return CustomManifest(components)



from typing import Any, Dict

from pysen.ext.flake8_wrapper import Flake8Setting


def test_flake8_setting_comment() -> None:
    flake8 = Flake8Setting().to_black_compatible()
    assert len(flake8._comments) > 0
    section_name, section = flake8.export()
    assert section_name == ["flake8"]
    comments: Dict[str, Any] = {k: v for k, v in section.items() if k.startswith("#")}
    assert len(comments) > 0
    # check that each comment entry doesn't have a value
    assert all(x is None for x in comments.values())
